from fastapi import APIRouter, Depends, File, UploadFile, HTTPException, Query
from starlette_context import context
import os
import logging
from requests.structures import CaseInsensitiveDict
from auth import auth
import json
import requests
import coloredlogs
router = APIRouter()
import conf
import _buildinfo as b
import opa
import correlation

_logger = logging.getLogger(__name__)
logger = correlation.DefaultExtrasAdapter(_logger, {"correlation_id": "None"})
#coloredlogs.install()
#logger = correlation.DefaultExtrasAdapter(_logger, {"correlation_id": context["correlation_id"]})

@router.get('/diag/require_authorized_user')
async def diag_require_authorized_user(auth_data: auth.Auth = Depends(auth.require_authorized_user)):
    """
    This API requires you to be an authorized user.
    Diagnostic API - For development testing only
    """
    return {"groups": auth_data.groups, "data_partition_id": auth_data.data_partition_id}

@router.get('/diag/require_authorized_admin')
async def diag_require_authorized_admin(auth_data: auth.Auth = Depends(auth.require_authorized_admin)):
    """
    This API requires you to be in authorized admin (be in admin group).
    Diagnostic API - For development testing only
    """
    return {"groups": auth_data.groups, "data_partition_id": auth_data.data_partition_id}

@router.get('/diag/require_data_partition_id')
async def diag_require_data_partition_id_in_header(data_partition_id: str = Depends(auth.require_data_partition_id)):
    """
    This testing API requires a data partition id in the header.
    Diagnostic API - For development testing only
    """
    return {'data_partition_id': data_partition_id}

@router.get("/diag/policies")
async def diag_fetch_all_policies(auth_data: auth.Auth = Depends(auth.require_authorized_admin)):
    """
    Return all policies from OPA directly with no filtering.
    Diagnostic API - For development testing only
    """
    logger = correlation.DefaultExtrasAdapter(_logger, {"correlation_id": context["correlation_id"]})
    opa_response = opa.fetch_opa_policies_direct()
    logger.debug(f"OPA return {opa_response.json}")
    if opa_response.ok:
        return opa_response.json
    else:
        raise HTTPException(status_code=opa_response.status_code, detail=opa_response.message)

@router.post("/diag/compile")
async def diag_compile(file: UploadFile,
    metrics: bool = Query(description='Include report detailed performance metrics on requested on individual API call. Returned inline with the API response', default=False),
    instrument: bool = Query(description='Include instrumentation data wth detailed performance metrics on requested on individual API call. Returned inline with the API response', default=False),
    auth_data: auth.Auth = Depends(auth.require_authorized_admin)):
    """
    # Test Compile.
    ## Diagnostic API - For development and performance testing only

    ### Metrics
    When query parameter metrics=true, the API response will include detailed performance metrics from OPA.
    OPA currently supports the following query performance metrics:

        timer_rego_input_parse_ns: time taken (in nanoseconds) to parse the input
        timer_rego_query_parse_ns: time taken (in nanonseconds) to parse the query.
        timer_rego_query_compile_ns: time taken (in nanonseconds) to compile the query.
        timer_rego_query_eval_ns: time taken (in nanonseconds) to evaluate the query.
        timer_rego_module_parse_ns: time taken (in nanoseconds) to parse the input policy module.
        timer_rego_module_compile_ns: time taken (in nanoseconds) to compile the loaded policy modules.
        timer_server_handler_ns: time take (in nanoseconds) to handle the API request.

    ### Instrumentation
    To enable query instrumentation, specify metrics=true and instrument=true query parameters when executing the API call.
    Query instrumentation can help diagnose performance problems, however, it can add significant overhead to query evaluation.
    We recommend leaving query instrumentation off unless you are debugging a performance problem.
    """
    logger = correlation.DefaultExtrasAdapter(_logger, {"correlation_id": context["correlation_id"]})
    contents = await file.read()
    data = contents.decode('utf-8')
    opa_response = opa.compile(query=data, metrics=metrics, instrument=instrument)
    logger.info(f"OPA compile return {opa_response.json}")
    if opa_response.ok:
        return opa_response.json
    else:
        raise HTTPException(status_code=opa_response.status_code, detail=opa_response.message)

@router.get('/diag/config')
async def diag_show_config(auth_data: auth.Auth = Depends(auth.require_authorized_admin)):
    """
    Return detail configuration details.
    Diagnostic API - For development testing only
    """
    logger = correlation.DefaultExtrasAdapter(_logger, {"correlation_id": context["correlation_id"]})
    r_status_json = {}
    r_config_json = {}
    r_config_headers = {}
    correlation_id = context["correlation_id"]
    user_agent = context["user_agent"]
    headers = CaseInsensitiveDict()
    headers["X-Request-ID"] = correlation_id
    headers["X-Correlation-ID"] = correlation_id
    headers["Correlation-ID"] = correlation_id
    headers["req_id"] = correlation_id
    try:
        #status_url = conf.OPA_STATUS_API + "/osdu/partition/" + auth_data.data_partition_id
        status_url = conf.OPA_STATUS_API 
        r_status = requests.get(status_url, timeout=10, headers=headers)
        r_status_json = r_status.json()
    except requests.exceptions.HTTPError:
        logger.error(f"Error: endpoint {status_url}: HTTPError")
    except:
        logger.error(f"Error: endpoint {status_url}: Error")

    try:
        r_config = requests.get(conf.OPA_CONFIG_API, timeout=10, headers=headers)
        r_config_json = r_config.json()
        r_config_headers = r_config.headers
    except requests.exceptions.HTTPError:
        logger.error(f"Error: endpoint {conf.OPA_CONFIG_API}: HTTPError")
    except:
        logger.error(f"Error: endpoint {conf.OPA_CONFIG_API}: Error")

    logger.info(f"config log test [{user_agent} {auth_data.x_user_id}]", extra={"correlation_id": context["correlation_id"]})
    return {
        "request": {
            "correlation_id": correlation_id,
            "user_agent": user_agent
        },
        "opa_config_response_headers": r_config_headers,
        "conf": {
            "ENTITLEMENTS_GROUPS_API": conf.ENTITLEMENTS_GROUPS_API,
            "LEGAL_BATCH_API": conf.LEGAL_BATCH_API,
            "ENABLE_DEV_DIAGNOSTICS": conf.ENABLE_DEV_DIAGNOSTICS,
            "OPA_URL": conf.OPA_URL,
            "OPA_POLICIES_API": conf.OPA_POLICIES_API,
            "OPA_DATA_API": conf.OPA_DATA_API,
            "OPA_COMPILE_API ": conf.OPA_COMPILE_API,
            "OPA_CONFIG_API": conf.OPA_CONFIG_API,
            "OPA_HEALTH_API": conf.OPA_HEALTH_API,
            "USE_BUNDLES": conf.USE_BUNDLES,
            "USER_PERMISION": conf.USER_PERMISION,
            "ADMIN_PERMISION": conf.ADMIN_PERMISION,
            "SERVICE_BASE_PATH": conf.SERVICE_BASE_PATH,
            "ENABLE_ADMIN_UI": conf.ENABLE_ADMIN_UI,
        },
        "env": {
            "CLOUD_PROVIDER": os.environ.get("CLOUD_PROVIDER"),
            "POLICY_BUCKET": os.environ.get("POLICY_BUCKET"),
            "CONTAINER_NAME": os.environ.get("CONTAINER_NAME"),
            "STORAGE_ACCOUNT": os.environ.get("STORAGE_ACCOUNT")
        },
        "auth_data": {
            "data_partition_id": auth_data.data_partition_id,
            "groups": auth_data.groups
        },
        "builddata": {
            "version": b.version,
            "artifactId": b.artifactId,
            "name": b.name,
            "groupId": b.groupId,
            "buildTime": b.buildTime,
            "branch": b.branch,
            "commitId": b.commitId,
            "commitMessage": b.commitMessage
        },
        "opa_status": r_status_json,
        "opa_config": r_config_json
    }
