# policy_api.py

# Copyright 2022 <XYZ>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import fastapi
from fastapi import Depends, HTTPException, UploadFile
import sys
import os
import logging
from starlette.requests import Request

from starlette.status import (
    HTTP_200_OK, 
    HTTP_400_BAD_REQUEST, 
    HTTP_422_UNPROCESSABLE_ENTITY,
    HTTP_501_NOT_IMPLEMENTED
)

from opa_response import OpaResponse
sys.path.append(os.path.abspath('..'))
import conf
import opa
from datetime import date
from auth import auth
from bundles import bundle

logger = logging.getLogger(__name__)

router = fastapi.APIRouter()

# Header code is compatible with Python 3.6 - 3.9, it will not work with Python 3.10
# See https://fastapi.tiangolo.com/tutorial/header-params/

@router.delete("/policies/osdu/partition/{data_partition}/{policy_id}")
async def delete_partition_policy(policy_id: str, data_partition: str, auth_data: auth.Auth = Depends(auth.require_authorized_admin)):
    """
    ## Delete a partition policy
    * This API requires admin privileges (service.policy.admin) in the provided data partition.
    * Partition ID in header and in path must match
    * Policy_id must end with a ".rego"

    ### Possible http return status codes:
    * 200 - no error
    * 202 - accepted
    * 400 - bad request
        * for example, if data_partition in path doesn't match data-partition-id in header
    * 401 - unauthorized
    * 403 - forbidden
        * for example, if calling with only user privs
    * 404 - not found
    * 422 - validation Error
        * for example, if policy_id doesn't end with ".rego"
    * 500 - server error
    * 501 - not implemented
        * for example, if bundles are not supported.
    * 503 - service not available
        * for example, if issues with bundles server

    Errors will include some detail in returning json.

    Return json:
    ```
        {
            "policy_id": string, 
            "data_partition": string,
            "status": bool,
            "message": string,
            "result": string json from OPA
        }
    ```
    """
    if conf.USE_BUNDLES is None:
        raise HTTPException(status_code=HTTP_501_NOT_IMPLEMENTED, detail="Partition specific policies are not supported on this OSDU instance")

    if data_partition != auth_data.data_partition_id:
        raise HTTPException(status_code=HTTP_400_BAD_REQUEST, detail="Provided data partition id does not correspond to provided policy id")

    if not policy_id.endswith(".rego"):
        raise HTTPException(status_code=HTTP_422_UNPROCESSABLE_ENTITY, detail="Policy ID should end with a .rego")

    cloud_provider = os.environ.get('CLOUD_PROVIDER')
    id = f"osdu/partition/{data_partition}/{policy_id}"
    if cloud_provider is None:
        # Help support MOCK testing
        result = OpaResponse()
        logger.critical(f"Error: CLOUD_PROVIDER ENV VAR not set / Mocking results for policyid: {id}")
        result.message = f"MOCK OK policy_id: {id}"
        result.ok = True
    elif cloud_provider == conf.MOCK:
        result = OpaResponse()
        logger.critical(f"Warning: CLOUD_PROVIDER ENV VAR set to Mock results for policyid: {id}")
        result.message = f"MOCK OK policy_id: {id}"
        result.ok = True
    elif cloud_provider == conf.LOCAL:
        # LOCAL Development ENV
        logger.critical(f"Error: CLOUD_PROVIDER ENV VAR LOCAL, Using local / direct OPA {conf.OPA_URL} policyid: {id}")
        result = opa.delete_opa_policy_direct(id)
    else:
        logger.debug(f"delete policy_id: {policy_id} id: {id}")
        result = bundle.delete_policy(data_partition=data_partition, policy_id=policy_id)

    if result.ok:
        return {
            "policy_id": policy_id, 
            "data_partition": data_partition,
            "status": result.ok,
            "message": result.message,
            "result": result.json
            }
    else:
        logger.error(f"Error when talking to OPA {result.message}")
        raise HTTPException(status_code=result.status_code, detail=f"Error when talking to OPA {result.message}")

@router.put("/policies/osdu/partition/{data_partition}/{policy_id}")
async def create_or_update_partition_policy(request: Request, policy_id: str, data_partition: str, file: UploadFile, auth_data: auth.Auth = Depends(auth.require_authorized_admin)):
    """

    ## Create or update a policy with given policy_id for a given data partition.
    This API will create/update policy definition with provided Rego expression included in file and assign it with provided id.

    * This API requires admin privileges (service.policy.admin) in the provided data partition.
    * Partition ID in header and in path must match
    * Policy_id must end with a ".rego"

    ### Possible http return status codes:
    * 200 - no error
    * 202 - accepted
    * 400 - bad request
        * for example, if data_partition in path doesn't match data-partition-id in header
    * 401 - unauthorized
    * 403 - forbidden
        * for example, if calling with only user privs
    * 422 - validation Error
        * for example, if policy_id doesn't end with ".rego"
        * for example, if package declaration issue
    * 500 - server error
    * 501 - not implemented
        * for example, if bundles are not supported.
    * 503 - service not available
        * for example, if issues with bundles server

    Errors will include some detail in returning json.

    Return json:
    ```
        {
            "policy_id": string,
            "data_partition": string,
            "opa_payload": string,
            "status_code": http status code
            "status": bool
            "message": string
        }
    ```
    """
    if conf.USE_BUNDLES is None:
        raise HTTPException(status_code=HTTP_501_NOT_IMPLEMENTED, detail="Partition specific policies are not supported on this OSDU instance")

    if auth_data.data_partition_id != data_partition:
        raise HTTPException(status_code=HTTP_400_BAD_REQUEST, detail="Provided data partition id does not correspond to provided policy id")

    if not policy_id.endswith(".rego"):
        raise HTTPException(status_code=HTTP_422_UNPROCESSABLE_ENTITY, detail="Policy ID should end with a .rego")

    contents = await file.read()

    if not bundle.package_name_ok(data_partition, policy_id, contents):
        raise HTTPException(status_code=HTTP_422_UNPROCESSABLE_ENTITY, detail="Package declaration should be: {0}".format(bundle.correct_package_declaration(data_partition, policy_id)))

    cloud_provider = os.environ.get('CLOUD_PROVIDER')
    id = f"osdu/partition/{data_partition}/{policy_id}"
    if cloud_provider is None:
        # Help support MOCK testing
        logger.critical(f"Error: CLOUD_PROVIDER ENV VAR not set / Mocking results for policyid: {id}")
        result = OpaResponse()
        result.ok = True
        result.status_code = HTTP_200_OK
        result.message = f"MOCK Policy {id} added to OPA"
    elif cloud_provider == conf.MOCK:
        logger.critical(f"Warning: CLOUD_PROVIDER ENV VAR set to Mock results for policyid: {id}")
        result = OpaResponse()
        result.ok = True
        result.status_code = HTTP_200_OK
        result.message = f"MOCK Policy {id} added to OPA"
    elif cloud_provider == conf.LOCAL:
        # LOCAL Development ENV
        logger.critical(f"Error: CLOUD_PROVIDER ENV VAR LOCAL, Using local / direct OPA {conf.OPA_URL} policyid: {id}")
        result = opa.put_opa_policy_direct(policy_id=id, data=contents.decode('utf-8'))
    else:
        logger.debug(f"put policy_id: {policy_id} id: {id}")
        result = bundle.put_policy(data_partition=data_partition, policy_id=policy_id, policy=contents)
    
    if result.ok:
        # TODO dev testing - fix return
        return {
            "policy_id": policy_id,
            "data_partition": data_partition,
            "opa_payload": contents,
            "status_code": result.status_code,
            "status": result.ok,
            "message": result.message}
    else:
        logger.error(f"Error when talking to CSP/OPA {result.message}")
        raise HTTPException(status_code=result.status_code, detail=f"Error when talking to OPA: {result.message}")
