from email import policy
import fastapi
from fastapi import HTTPException, Depends, UploadFile
import requests
import json
import logging
router = fastapi.APIRouter()
import opa
import conf
from starlette_context import context
import correlation

from starlette.status import (
    HTTP_500_INTERNAL_SERVER_ERROR,
    HTTP_404_NOT_FOUND,
    HTTP_406_NOT_ACCEPTABLE,
    HTTP_422_UNPROCESSABLE_ENTITY
)

from models.translate import TranslateItem
from auth import auth
import translate.convert
from translate.translate_exceptions import PolicyDenyError, PolicyNotFoundError

_logger = logging.getLogger(__name__)
logger = correlation.DefaultExtrasAdapter(_logger, {"correlation_id": "None"})

@router.post("/translate", tags=["translate"])
async def translate_policy(trans:  TranslateItem,  auth_data: auth.Auth = Depends(auth.require_authorized_user)):
    """
    ## Translate policy
    Given an OPA query that should be partially evaluated, return an ElasticSearch request body

    In the body of the request the JSON schema should match "TranslateItem".
    Please note: xuserid, token and datapartitionid are now actively inserted into input request
    """
    logger = correlation.DefaultExtrasAdapter(_logger, {"correlation_id": context["correlation_id"]})
    posted_data = json.loads(trans.tojson()) # convert to dict for easy manipulation
    logger.debug(f"data: {posted_data}")

    if not 'input' in posted_data:
        raise HTTPException(status_code=HTTP_422_UNPROCESSABLE_ENTITY, detail=f"Input not found in data")

    posted_data["input"]["token"] = auth_data.access_token
    posted_data["input"]["xuserid"] = auth_data.x_user_id
    posted_data["input"]["datapartitionid"] = auth_data.data_partition_id
    logger.debug(f"data after adding auth: {posted_data}")

    try:
        r = requests.post(conf.OPA_COMPILE_API,
            data=json.dumps(posted_data), # OPA needs string here
            timeout=20,
            headers={'Content-Type': 'application/json'})
    except Exception as err:
        logger.error(err)
        raise HTTPException(status_code=HTTP_500_INTERNAL_SERVER_ERROR, detail="An error occurred when talking to OPA")

    if r.ok:
        js = r.json()
        logger.debug(f"OPA response: {js}")
        try:
            if "queries" in r.text:
                opa_ast = r.json()["result"]["queries"]
                logger.debug(f"OK opa response {opa_ast}")
            else:
                logger.info(f"No queries in opa response {js}")
                if policy_exists(query=posted_data, data_partition=auth_data.data_partition_id):
                    # When not getting a valid response from OPA, we want to still make search happy
                    # If policy exists return a false match response
                    # If policy does not exist return a HTTP_404_NOT_FOUND
                    return {
                        "query": {
                            "match_none": {}
                            }
                        }
                else:
                    raise PolicyNotFoundError(f'No query in opa response. Translate failed. Policy may not exist. {js}')

            es_request_body = translate.convert.convert_opa_ast_json_to_es_request_body(
                opa_ast, js
            )
            return es_request_body
        except PolicyNotFoundError as err:
            logger.error(err)
            raise HTTPException(status_code=HTTP_404_NOT_FOUND, detail=f"{err}")
        except PolicyDenyError as err:
            logger.error(err)
            raise HTTPException(status_code=HTTP_406_NOT_ACCEPTABLE, detail=f"An error occurred when talking translate service. {err}")
            #return {
            #    "query": {
            #        "must_not": {}
            #        }
            #    }
        except NotImplementedError as err:
            logger.error(err)
            raise HTTPException(status_code=HTTP_406_NOT_ACCEPTABLE, detail=f"An error occurred when talking translate service. {err}")
        except KeyError as err:
            logger.error(err)
            raise HTTPException(status_code=HTTP_406_NOT_ACCEPTABLE, detail=f"An error occurred when talking translate service. {err}")
        except Exception as err:
            logger.error(err)
            template = "An exception of type {0} occurred. Arguments:\n{1!r}"
            message = template.format(type(err).__name__, err.args)
            logger.error(message)
            raise HTTPException(status_code=HTTP_500_INTERNAL_SERVER_ERROR, detail=f"An error occurred when talking translate service. {err} {message}")
    else:
        logger.error(f"Error from OPA: {r.text}")
        raise HTTPException(status_code=HTTP_500_INTERNAL_SERVER_ERROR, detail=f"OPA Error: {r.text}")

def policy_exists(query, data_partition):
    if "query" in query:
        p = query["query"]
        if p.startswith('data.osdu.partition['):
            policy_id = p.split('.')[3] + ".rego"
            opa_response = opa.fetch_opa_policy_direct(conf.PARTITION_BUNDLE_ROOT.format(data_partition) + "/{0}".format(policy_id))
        elif p.startswith('data.osdu.instance.'):
            policy_id = p.split('.')[3] + ".rego"
            opa_response = opa.fetch_opa_policy_direct(conf.INSTANCE_BUNDLE_ROOT+"/{0}".format(policy_id))
        else:
            logger.info(f"policy_exists unexpected query {p}")
            return False
        return opa_response.ok
