from os import access
from fastapi import Depends, Request, Header, HTTPException
from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer
from requests import head
from starlette.authentication import AuthCredentials
from requests.structures import CaseInsensitiveDict
import requests
import logging
import json
import conf
from starlette_context import context
from entitlement import EntitlementService

from starlette.status import (
    HTTP_200_OK, 
    HTTP_201_CREATED, 
    HTTP_400_BAD_REQUEST, 
    HTTP_401_UNAUTHORIZED, 
    HTTP_403_FORBIDDEN,
    HTTP_404_NOT_FOUND, 
    HTTP_422_UNPROCESSABLE_ENTITY,
)

logger = logging.getLogger(__name__)

security = HTTPBearer()

async def require_token(request: Request,
                                          credentials: HTTPAuthorizationCredentials = Depends(security)):
    token = credentials.credentials
    request.scope['auth'] = AuthCredentials(['authenticated'])
    return token

async def require_data_partition_id(
        data_partition_id: str = Header(default=None,
                                        title='data partition id',
                                        description='identifier of the data partition to query',
                                        min_length=1)):
    if data_partition_id is None:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=f"Bad Request. Missing data_partition_id in header",
            headers={"WWW-Authenticate": "Bearer"},
        )
    return data_partition_id

class Auth:
    def __init__(self, access_token, data_partition_id = None, x_user_id = None, authenticated = False, entitle_service = EntitlementService):
    #def __init__(self, access_token, data_partition_id = None, x_user_id = None, authenticated = False):
        self.access_token = access_token
        self.data_partition_id = data_partition_id
        self.x_user_id = x_user_id
        self.authenticated = authenticated
        self.groups = []
        self.entitle_service = entitle_service

    def __repr__(self): 
        return f"AUTH access_token: {self.access_token}\ndata_partition_id: {self.data_partition_id}\nx_user_id: {self.x_user_id}\ngroups: {self.groups}"

async def require_authorized_user(
        request: Request, credentials: HTTPAuthorizationCredentials = Depends(security),
        data_partition_id: str = Header(default=None,
            description='identifier of the data partition to query',
            min_length=1),
        x_user_id: str = Header(default=None, 
            description='identifier the user in the query',
            min_length=1)):

    token = credentials.credentials.strip()

    auth_data = Auth(access_token=token, data_partition_id=data_partition_id, x_user_id=x_user_id)
    headers = CaseInsensitiveDict()
    headers["Authorization"] = "Bearer " + token
    headers["data-partition-id"] = data_partition_id
    headers["content-type"] = "application/json"
    if context["correlation_id"]:
        headers["correlation_id"] = context["correlation_id"]

    groups = auth_data.entitle_service.request_groups(headers = headers)
    logger.debug(f"response from entitlement service for {data_partition_id}: {groups}")
    auth_data.authenticated = True

    # add groups to auth_data
    try:
        #for val in response.json()['groups']:
        for val in groups:
            if isinstance(val,dict):
                auth_data.groups.append(val['name'])
    except Exception:
        pass

    #print("Auth data groups:", auth_data.groups)

    # if user is not in USER_PERMISSION group and not and not in ADMIN group
    if conf.USER_PERMISION not in auth_data.groups and conf.ADMIN_PERMISION not in auth_data.groups:
        raise HTTPException(
            status_code=HTTP_401_UNAUTHORIZED,
            detail="User does not have the permission to call the API",
            headers={"WWW-Authenticate": "Bearer"},
        )
    logger.debug(f"Auth data OK {data_partition_id}")
    return auth_data

async def require_authorized_admin(request: Request, auth_data: Auth = Depends(require_authorized_user)):
    
    if conf.ADMIN_PERMISION not in auth_data.groups:
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail="User does not have the permission to call the Admin API",
            headers={"WWW-Authenticate": "Bearer"},
        )
    return auth_data
