import logging
from os import environ
import tarfile
import io
from fastapi import HTTPException

from starlette.status import (
    HTTP_200_OK, 
    HTTP_201_CREATED, 
    HTTP_202_ACCEPTED,
    HTTP_400_BAD_REQUEST, 
    HTTP_401_UNAUTHORIZED, 
    HTTP_403_FORBIDDEN,
    HTTP_404_NOT_FOUND, 
    HTTP_422_UNPROCESSABLE_ENTITY,
    HTTP_500_INTERNAL_SERVER_ERROR,
    HTTP_502_BAD_GATEWAY,
    HTTP_503_SERVICE_UNAVAILABLE
)

import conf
from bundles import storage
from osdu_api.providers.types import FileLikeObject

from opa_response import OpaResponse

logger = logging.getLogger(__name__)

def delete_policy(data_partition, policy_id):
    """
    Delete policy from CSP providers
    """

    response = OpaResponse()
    response.message = f"Unable to remove policy {policy_id} from bundle server"

    # Double-check to make sure we have a .rego
    if not policy_id.endswith(".rego"):
        return response

    found_policy_in_bundle = False
    bundle = bundle_filename(data_partition)
    with io.BytesIO() as outbuffer:
        with io.BytesIO() as inbuffer:
            try:
                inbuffer, uri = storage.get_storage().download_file(bundle, inbuffer)
            except Exception as err:
                logger.error(f"Error Downloading {bundle} from CSP Storage: {err}")
                response.status_code = HTTP_503_SERVICE_UNAVAILABLE
                return response
            inbuffer.seek(0)
            with tarfile.open(mode="w:gz", fileobj=outbuffer) as tarout:
                with tarfile.open(mode='r:gz',fileobj=inbuffer) as tarin:
                    for member in tarin.getmembers():
                        # Skip policy to be deleted
                        if member.name == policy_id:
                            found_policy_in_bundle = True
                        else:
                            tarout.addfile(member, tarin.extractfile(member.name))
                tarout.close()
                outbuffer.seek(0)
                if found_policy_in_bundle:
                    if storage.get_storage().upload_file(bundle, outbuffer):
                        response.message = f"Policy {policy_id} removed from bundle server"
                        # Could be 200, but 202 is proper since there could be OPA servers that may still have the content
                        response.status_code = HTTP_202_ACCEPTED # could be 200
                        response.ok = True
                        return response
                else:
                    # Return NOT FOUND, no need to upload bundle
                    # Text should contain "not_found" as OPA includes that when talking directly
                    response.message = f"Policy {policy_id} not_found in bundle server"
                    response.status_code = HTTP_404_NOT_FOUND
    return response

def put_policy(data_partition, policy_id, policy):
    """
    Put policy as bundle to CSP providers
    """
    bundle = bundle_filename(data_partition)
    logger.debug(f"bundle: {bundle}")

    response = OpaResponse()
    response.message = f"Unable to add policy {policy_id} to bundle server"

    with io.BytesIO() as outbuffer:
        with io.BytesIO() as inbuffer:
            try:
                inbuffer, uri = storage.get_storage().download_file(bundle, inbuffer)
            except Exception as err:
                logger.critical(f"Error Downloading {bundle} from CSP Storage: {err}")
                response.status_code = HTTP_503_SERVICE_UNAVAILABLE
                template = "An exception of type {0} occurred. Arguments:\n{1!r}"
                message = template.format(type(err).__name__, err.args)
                logger.critical(message)
                return response
            inbuffer.seek(0)
            with tarfile.open(mode="w:gz", fileobj=outbuffer) as tarout:
                with tarfile.open(mode='r:gz',fileobj=inbuffer) as tarin:
                    for member in tarin.getmembers():
                        # Do not add previous version of existing policy (in case it existed)
                        if member.name!= policy_id:
                            tarout.addfile(member, tarin.extractfile(member.name))
                info = tarfile.TarInfo(policy_id)
                info.size = len(policy)
                tarout.addfile(info, io.BytesIO(policy))
                tarout.close()
                outbuffer.seek(0)
                if storage.get_storage().upload_file(bundle, outbuffer):
                    response.ok = True
                    response.message = f"Policy {policy_id} added to bundle server"
                    # could be 200, but 202 is proper since there could be OPA servers that may not have the content yet
                    response.status_code = HTTP_202_ACCEPTED
                    return response
                else:
                    logger.critical(f"Error uploading {bundle} to CSP Storage")

    return response

def policy_filename(policy_id):
    return '{0}.rego'.format(policy_id)

def bundle_filename(data_partition):
    if data_partition:
        logger.debug(f"partition {data_partition} bundle")
        return f"bundle-{data_partition}.tar.gz"
    else:
        logger.debug("Instance/default bundle")
        return 'bundle.tar.gz'

def correct_package_declaration(data_partition, policy_id):
    return 'package osdu.partition["{0}"].{1}'.format(data_partition, policy_id[:-5])

def package_name_ok(data_partition, policy_id, policy):
    logger.debug(f"data_partition: {data_partition} policy_id: {policy_id}")
    return correct_package_declaration(data_partition, policy_id) in policy.decode("utf-8")
