#!/usr/bin/env python3
# OSDU Policy Service
from fastapi import FastAPI, Request, Depends, BackgroundTasks, Header, Depends, Response
from fastapi.responses import RedirectResponse, HTMLResponse
from fastapi.security import OAuth2PasswordBearer, OAuth2PasswordRequestForm, HTTPBearer
from fastapi.openapi.docs import get_swagger_ui_html
from fastapi.openapi.utils import get_openapi
from fastapi.staticfiles import StaticFiles
from starlette_context import context, request_cycle_context
import uvicorn
import os
import coloredlogs
import logging
import conf
from api import policy_read_api, policy_update_api, policy_eval_api, health_api, info_api, translate_api, diag_api, login_api
from views import home
import _buildinfo as b
from auth import auth
#from fastapi.middleware.cors import CORSMiddleware
from starlette.middleware import Middleware
from starlette.middleware.cors import CORSMiddleware
import correlation 

tags_metadata = [
    {
        "name": "policies",
        "description": "Policy Maintenance"
    },
    {
        "name": "evaluate",
        "description": "Evaluate APIs"
    },
    {
        "name": "translate",
        "description": "Translation APIs"
    },
    {
        "name": "diagnostic",
        "description": "Version information and other diagnostic APIs"
    },
    {
        "name": "login",
        "description": "Login for Admin UI"
    },
    {
        "name": "home",
        "description": "Home Page"
    }
]

description = f"""
    OSDU Policy Service v{b.version} API 

Policy service is used for management and evaluation of dynamic policies in OSDU.

### Dependencies:
Policy service has a runtime dependency on the following:
* Open Policy Agent [OPA](https://openpolicyagent.org/)
* OSDU entitlements service for API authorization

### API authorization
* To read or evaluate policies, user calling the policy service, must be a member of _service.policy.user_ or _service.policy.admin_ group.
* To create, update, or delete the policies, user must be a member of _service.policy.admin_ group.
* User can be added to these groups by using entitlements service.
"""

# setup loggers and make sure we can find logging.config
log_file_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'logging.conf')
logging.config.fileConfig(log_file_path, disable_existing_loggers=False)

# get root logger
logger = logging.getLogger(__name__)  # the __name__ resolve to "main" since we are at the root of the project. 
                                       # This will get the root logger since no logger in the configuration has this name.
#logger = correlation.DefaultExtrasAdapter(_logger, {"correlation_id": "None"})
#coloredlogs.install()     # Add colored logs support

#old_factory = logging.getLogRecordFactory()
#
#def record_factory(*args, **kwargs):
#    record = old_factory(*args, **kwargs)
#    record.correlation_id = "None"
#    return record
#
#logging.setLogRecordFactory(record_factory)

# disable docs
# app = FastAPI(docs_url=None, redoc_url=None, openapi_url = None)

middleware = [
    Middleware(
        CORSMiddleware,
        allow_origins=['*'],
        allow_credentials=True,
        allow_methods=['*'],
        allow_headers=['*']
    )
]

"""
async def policy_context_dependency(correlation_id: str = Header(None), user_agent: str = Header(None)):
    # When used a Depends(), this fucntion get the `Correlation-ID` header,
    # which will be documented as a required header by FastAPI.
    # use `correlation_id: str = Header(None)` for an optional header.
    # (correlation_id = Header(...) for required header

    data = {"correlation_id": correlation_id,
            "user_agent": user_agent
           }
    with request_cycle_context(data):
        # yield allows it to pass along to the rest of the request
        yield
"""

app = FastAPI(
    middleware=middleware,
    title="Policy Service",
    description=description,
    version=b.version,
    docs_url=conf.SERVICE_BASE_PATH + "/docs",
    redoc_url=conf.SERVICE_BASE_PATH + "/redoc",
    openapi_url=conf.SERVICE_BASE_PATH + "/openapi.json",
    contact={
        "name": "OSDU Policy Service",
        "url": "https://community.opengroup.org/osdu/platform/security-and-compliance/policy"},
    license_info={
        "name": "Apache 2.0",
        "url": "https://community.opengroup.org/osdu/platform/security-and-compliance/policy/-/raw/trusted-FastAPI/LICENSE",
    },
    dependencies=[Depends(correlation.policy_context_dependency)],
    openapi_tags=tags_metadata)

#@app.on_event("startup")
#async def startup_event():
#    logger = logging.getLogger("uvicorn.access")
#    handler = logging.StreamHandler()
#    handler.setFormatter(logging.Formatter("%(asctime)s - %(levelname)s - %(message)s"))
#    logger.addHandler(handler)

def configure():
    
    configure_routing()
    if conf.ENABLE_ADMIN_UI:
        # only needed to support Admin UI
        logger.info("AdminUI enabled")
        #configure_middleware()
        app.mount("/adminui", StaticFiles(directory="assets", html=True), name="assets")


def configure_routing():
    ADMIN_PROTECTED = [Depends(auth.require_authorized_admin)]
    PROTECTED = [Depends(auth.require_authorized_user)]

    app.include_router(home.router, tags=["home"])
    app.include_router(policy_read_api.router, prefix=conf.SERVICE_BASE_PATH, tags=["policies"], dependencies=PROTECTED)
    app.include_router(policy_update_api.router, prefix=conf.SERVICE_BASE_PATH, tags=["policies"], dependencies=ADMIN_PROTECTED)
    app.include_router(policy_eval_api.router, prefix=conf.SERVICE_BASE_PATH, tags=["evaluate"], dependencies=PROTECTED)
    app.include_router(translate_api.router, prefix=conf.SERVICE_BASE_PATH, tags=["translate"], dependencies=PROTECTED)
    app.include_router(info_api.router, prefix=conf.SERVICE_BASE_PATH, tags=["diagnostic"], dependencies=PROTECTED)
    app.include_router(health_api.router, prefix=conf.SERVICE_BASE_PATH, tags=["diagnostic"])

    if conf.ENABLE_ADMIN_UI:
        # only needed to support Admin UI
        app.include_router(login_api.router, prefix=conf.SERVICE_BASE_PATH, tags=["login"], dependencies=ADMIN_PROTECTED)

    if conf.ENABLE_DEV_DIAGNOSTICS:
        # do not enable diags in production
        app.include_router(diag_api.router, tags=["diagnostic"], dependencies=ADMIN_PROTECTED)
        logger.warning("ENABLE_DEV_DIAGNOSTICS enabled")

def configure_middleware():
    # This will need some work before Admin UI is moved out of POC
    origins = [
        "http://localhost",
        "http://localhost:8080",
        "http://0.0.0.0:8080",
        "http://localhost:4200"
    ]
    app.add_middleware(
        CORSMiddleware,
        allow_origins=['*'],
        allow_credentials=True,
        allow_methods=['*'],
        allow_headers=['*']
    )

if __name__ == '__main__':
    configure()
    #log_config = uvicorn.config.LOGGING_CONFIG
    #log_config["formatters"]["access"]["fmt"] = "%(asctime)s - %(levelname)s - %(message)s"
    #log_config["formatters"]["default"]["fmt"] = "%(asctime)s - %(levelname)s - %(message)s"
    uvicorn.run(app, port=80, host='0.0.0.0', log_config=None)
else:
    configure()
