from urllib import response
import conf
import requests
import logging
import json
from opa_response import OpaResponse

logger = logging.getLogger(__name__)

def fetch_opa_policies_direct(timeout=20):
    """ fetch a list of all policies in OPA """
    result = OpaResponse()
    try:
        rsp = requests.get(conf.OPA_POLICIES_API, timeout=timeout)
        result.status_code = rsp.status_code
        result.ok = rsp.ok
        result.json = rsp.json()
        if rsp.ok:
            #logging.info(f"result from OPA ({conf.OPA_POLICIES_API}): {json.dumps(result.json, indent=2)}")
            logger.debug(f"result from OPA ({conf.OPA_POLICIES_API}) OK")
        else:
            logger.error("Error fetching policies, got status %s and message: %s" % (rsp.status_code, rsp.text))
            if "message" in result.json:
                result.message = result.json["message"]
            elif "code" in result.json:
                result.message = result.json["code"]
    except Exception as err:
        logger.error(err)
    return result

def fetch_opa_policy_direct(policy_id, timeout=20):
    """
    fetch a specific policy in OPA
    """
    url = conf.OPA_POLICIES_API + '/' + policy_id
    
    logger.debug(f"OPA get URL: {url}")
    result = OpaResponse()
    try:
        rsp = requests.get(url, timeout=timeout)
        result.status_code = rsp.status_code
        result.ok = rsp.ok
        result.json = rsp.json()
        if rsp.ok:
            #logging.debug(json.dumps(result.json, indent=2))
            logger.debug(f"get {url} OK")
        else:
            logger.error(f"Error fetching policies from {url}, got status {rsp.status_code} and message: {rsp.text}")
            if "message" in result.json:
                result.message = result.json["message"]
            elif "code" in result.json:
                result.message = result.json["code"]

    except Exception as err:
        logger.error(err)
    return result

def delete_opa_policy_direct(policy_id, timeout=20):
    """
    Delete a policy directly from OPA
    """
    url = conf.OPA_POLICIES_API + '/' + policy_id
    result = OpaResponse()
    result.message = f"Error while removing policy {policy_id}"

    try:
        rsp = requests.delete(url, timeout=timeout)
        result.status_code = rsp.status_code
        result.ok = rsp.ok
        result.json = rsp.json()
        if rsp.ok:
            logger.info(f"deleted policy {policy_id} from {url}, and message: {rsp.text}")
            logger.debug(json.dumps(result.json, indent=2))
            if result.json == {}:
                result.message =f"Policy {policy_id} removed from OPA server"
            else:
                result.message =f"Policy {policy_id} removed from OPA server, but expected return {result.json}"
        else:
            logger.error(f"Error deleting policy {policy_id} from {url}, got status {rsp.status_code} and message: {rsp.text}")
            if "message" in result.json:
                result.message = result.json["message"]
            elif "code" in result.json:
                result.message = result.json["code"]

    except Exception as err:
        logger.error(err)
    return result

def put_opa_policy_direct(policy_id, data, timeout=20):
    """
    Put a policy in OPA directly for use when not using a CLOUD_PROVIDER
    """
    url = conf.OPA_POLICIES_API + '/' + policy_id
    logger.debug(f"OPA put URL {url}")
    result = OpaResponse()
    try:
        rsp = requests.put(url, data=data, headers={'Content-Type': 'application/octet-stream'}, timeout=timeout)
        result.status_code = rsp.status_code
        result.ok = rsp.ok
        result.json = rsp.json()
        if rsp.ok:
            logger.debug(f"Response from OPA({url}): {json.dumps(result.json, indent=2)}")
            result.message = f"Policy {policy_id} added to OPA"
        else:
            logger.error("Error updating policy, got status %s and message: %s" % (rsp.status_code, rsp.text))
            result.message = f"Error updating policy: {rsp.text}"
    except Exception as err:
        logger.error(f"Unexpected error when attempting to connect to OPA: {err}")
        result.message = "Unexpected error"
    return result

def compile(query, metrics=False, instrument=False, timeout=20):
    """
    Compile - Partially Evaluate a Query.
    The Compile API allows you to partially evaluate Rego queries and obtain a simplified version of the policy.  

    OPA can report detailed performance metrics at runtime.
    Performance metrics can be requested on individual API calls and are returned inline with the API response.
    To enable performance metric collection on an API call, specify the metrics=true query parameter when executing the API call.

    OPA also supports query instrumentation.
    To enable query instrumentation, specify the instrument=true query parameter when executing the API call.
    Query instrumentation can help diagnose performance problems, however, it can add significant overhead to query evaluation.
    We recommend leaving query instrumentation off unless you are debugging a performance problem.
    """
    url = conf.OPA_COMPILE_API
    if metrics:
        url = url + '?metrics=true'

    if instrument and metrics:
        url = url + '&instrument=true'

    logger.debug(f"OPA compile URL {url} type: {type(query)} json_data: {query}")
    result = OpaResponse()

    try:
        rsp = requests.post(url,
            data=query,
            timeout=timeout,
            headers={'Content-Type': 'application/json'})

        result.status_code = rsp.status_code
        result.ok = rsp.ok
        result.json = rsp.json()

        if rsp.ok:
            logger.debug(f"OK Response from OPA({url}): {json.dumps(result.json, indent=2)}")
            result.message = f"Policy OPA compiled"
        else:
            logger.error("Error compiling policy, got status %s and message: %s" % (rsp.status_code, rsp.text))
            result.message = f"Error compiling policy: {rsp.text}"
        
    except Exception as err:
        logger.error(err)
    return result

def data(query, path, timeout=20):
    """
    Compile - Partially Evaluate a Query.
    The Compile API allows you to partially evaluate Rego queries and obtain a simplified version of the policy.  
    """
    url = conf.OPA_DATA_API + "/" + path
    logger.debug(f"OPA compile URL {url} type: {type(query)} json_data: {query}")
    result = OpaResponse()

    try:
        rsp = requests.post(url,
            data=query,
            timeout=timeout,
            headers={'Content-Type': 'application/json'})

        result.status_code = rsp.status_code
        result.ok = rsp.ok
        result.json = rsp.json()

        if rsp.ok:
            logger.debug(f"OK Response from OPA({url}): {json.dumps(result.json, indent=2)}")
            result.message = f"Policy OPA query"
        else:
            logger.error("Error compiling policy, got status %s and message: %s" % (rsp.status_code, rsp.text))
            result.message = f"Error query policy: {rsp.text}"
        
    except Exception as err:
        logger.error(err)
    return result

def get_document_with_input(rego_package_name, rego_rule_name, rego_input, timeout=20):
    """
    Get data from OPA
    """
    url = conf.OPA_DATA_API + "/" + rego_package_name + "/" + rego_rule_name
    logger(f"get_document_with_input url: {url}")
    response = requests.post(url, data=json.dumps({"input": rego_input}), timeout=timeout)
    #print(rego_input)
    #print(response.json())
    result = response.json()["result"]
    return result