# conftest.py
import pytest
import os

"""
Test Config Notes:

See README.md in testing directory for details and requirements for running tests.

TLDR:
* TOKEN or --token
* CLOUD_PROVIDER 
* DOMAIN or --domain
* DATA_PARTITION or --data_partition
* --service_url
* --opa_access (if pytest can talk to OPA)
* BUNDLE_PAUSE or --bundle_pause (make sure this is 2-3x your max polling for OPA will default to 120 seconds)

In addition Environ Variables used in conf.py must be set: 
* ENTITLEMENTS_BASE_URL
* LEGAL_BASE_URL
* OPA_URL
"""

def pytest_addoption(parser):
    parser.addoption("--opa_access", action="store_true", default=False, help="access to OPA from pytest allowed")
    parser.addoption("--domain", action="store", help="Specify domain to use")
    parser.addoption("--token", action="store", help="Specify token to use")
    parser.addoption("--service_url", action="store", help="policy service endpoint base url", default="http://localhost:8080")
    parser.addoption("--bundle_pause", action="store", help="how long to wait for bundles sync")
    parser.addoption("--data_partition", action="store", help="data-partition-id to use")

@pytest.fixture(scope='session')
def token(request):
    token_val = request.config.option.token
    if token_val is None:
        token_val = os.getenv('TOKEN')
    if token_val is None:
        print("TOKEN is required.")
    return token_val

@pytest.fixture(scope='session')
def opa_access(request):
    return request.config.option.opa_access

@pytest.fixture(scope='session')
def bundle_pause(request):
    if request.config.option.bundle_pause is None:
        pause_val = os.getenv('BUNDLE_PAUSE')
        if pause_val is None:
            pause_val = 120
    else:
        pause_val = int(request.config.option.bundle_pause)
    return int(pause_val)

@pytest.fixture
def cloud_provider():
    cloud = os.environ.get('CLOUD_PROVIDER')
    if cloud is None:
        print("CLOUD_PROVIDER is required.")
    return cloud

@pytest.fixture(scope='session')
def service_url(request):
    return request.config.option.service_url

@pytest.fixture(scope='session')
def domain(request):
    domain_val = request.config.option.domain
    if domain_val is None:
        domain_val = os.getenv('DOMAIN')
    if domain_val is None:
        print("DOMAIN is required.")
    return domain_val

@pytest.fixture(scope='session')
def data_partition(request):
    data_partition_val = request.config.option.data_partition
    if data_partition_val is None:
       data_partition_val = os.getenv('DATA_PARTITION')
    if data_partition_val is None:
        data_partition_val = "osdu"
    return data_partition_val

def pytest_configure(config):
    config.addinivalue_line("markers", "opa_access: mark test as connects to OPA")

def pytest_collection_modifyitems(config, items):
    if config.getoption("--opa_access"):
         # --opa_access given in cli: do not skip tests that connect to OPA
        return
    skip_opa_access = pytest.mark.skip(reason="need --opa_access option to run")
    for item in items:
        if "opa_access" in item.keywords:
            item.add_marker(skip_opa_access)

# def pytest_collection_modifyitems(items):
#     """Modifies test items in place to ensure test modules run in a given order."""
#     MODULE_ORDER = ["tests.test_b", "tests.test_c", "tests.test_a"]
#     print(items)
#     module_mapping = {item: item.module.__name__ for item in items}

#     sorted_items = items.copy()
#     # Iteratively move tests of each module to the end of the test queue
#     for module in MODULE_ORDER:
#         sorted_items = [it for it in sorted_items if module_mapping[it] != module] + [
#             it for it in sorted_items if module_mapping[it] == module
#         ]
#     items[:] = sorted_items
