# test_integration_035_eval.py
import pytest
# requires pytest_dependency to be installed
import unittest
import responses
from fastapi.testclient import TestClient
import requests
import re
import logging
import sys
import os
import base64
import json
import random
from pathlib import Path
from string import Template

# so the directly in which you run pytest or unittest doesn't matter as much
sys.path.append(os.path.abspath('..'))
sys.path.append('tests')

# local
from main import app
from auth import auth
import conf
import common

# override dependency injection for authentication to entitlement service
from override_depends import override_require_authorized_user, set_authorize_session, ADMIN_ONLY_SVC, USER_AND_ADMIN_SVC, USER_ONLY_SVC, OTHER_ONLY_SVC

TEST_DATA_DIR = Path(__file__).resolve().parent / 'templates'

client = TestClient(app)

@pytest.mark.dependency(name="require_token")
def test_require_token(token):
    assert token is not None, "No token provided on command line"

def post_query(policy_id, token, data_partition, files, service_url = False, expect=200):
    url = conf.SERVICE_BASE_PATH + "/evaluations/query"
    params = {'include_auth': True, 'policy_id': policy_id}
    if service_url:
        r = requests.post(service_url + url,
            files = files,
            params=params,
            headers={'Authorization': 'Bearer ' + token,
            'data-partition-id': data_partition})
    else:
        r = client.post(url,
            files = files,
            params=params,
            headers={'Authorization': 'Bearer ' + token,
            'data-partition-id': data_partition})

    if not r.ok and expect == 200:
        json_str = json.dumps(r.json(), indent=4, sort_keys=True)
        filename = "pytest_eval_query_debug.json"
        with open(filename, "w") as outfile:
            outfile.write(json_str)
            outfile.close()
            print(f"NOTICE: {filename} saved")
    assert r.status_code == expect, f"Expected to get {expect} from {url} {r.text}"
    result = r.json()
    if expect == 200:
        assert "result" in r.text, f"Expected to get 'result' from {url} {r.text}"
        assert "records" in r.text, f"Expected to get 'records' from {url} {r.text}"
        assert "errors" in r.text, f"Expected to get 'errors' from {url} {r.text}"
        assert "id" in r.text, f"Expected to get 'id' from {url} {r.text}"
        print(result["result"]["records"])
        errors = result["result"]["records"][0]["errors"]
        print(f"errors: {errors}")
        if not errors == []:
            print("Found Errors: Check to make sure id, legal tag and domain are valid!")
            knownissue=[{'code': '403', 'id': 'osdu:test:1.4.1654807204111', 'message': 'The user is not authorized to perform this action', 'reason': 'Access denied'}]
            if errors == knownissue:
                pytest.skip("Skipping eval test - misconfigured id in eval, otherwise working")
        assert errors == [], f"Expected no OPA errors from {url} {r.text}"
    return result

@pytest.mark.dependency(depends=["require_token"])
def test_eval_dataauthz(token, data_partition, service_url, domain):
    eval_policy(token=token,
        data_partition=data_partition,
        service_url=service_url,
        policy_name="dataauthz",
        domain=domain)

@pytest.mark.dependency(depends=["require_token"])
def test_eval_doesnotexist(token, data_partition, service_url, opa_access, domain):
    eval_policy(
        token=token,
        data_partition=data_partition,
        service_url=service_url,
        policy_name="foobar",
        expect=406,
        opa_access=opa_access,
        domain=domain)

@pytest.mark.dependency(depends=["require_token"])
def eval_policy(token, data_partition, service_url, policy_name, expect=200, opa_access=False, domain="example.com"):
    """
    Test Evaluation reading in tests/template/eval*.json
    * This requires dataauthz.rego to be already loaded in the data partition
    * Replace ${data_partition} with data partition id
    * Replace ${legal_tag} with a randomly selected legal tag that is valid with that data partition,
    * Replace ${name} with policy name
    * Replace ${domain} with domain name
        retrieved from legal service
    """
    #if conf.ENABLE_DEV_DIAGNOSTICS == False:
    #    pytest.skip("Skipping diagnostic test")
    assert token is not None, "No token provided on command line"
    assert data_partition is not None, "No data partition provided on command line"

    policy_id = "osdu/partition/" + data_partition + "/" + policy_name

    # randomly select a legal tag
    legal_tag = random.choice(get_legal_tags(token=token, data_partition=data_partition))
    print(f"Using Legal tag: {legal_tag}")

    p = os.path.dirname(os.path.abspath(__file__))
    template_datadir = os.path.join(p, '..', 'templates')
    for filename in sorted(os.listdir(template_datadir)):
        filetype='.json'
        file_path = os.path.join(template_datadir, filename)
        # checking if it is a file
        if os.path.isfile(file_path):
            if file_path.endswith(filetype) and filename.startswith("eval"): # eval*.json
                input_json = os.path.basename(file_path)
                print(input_json)
                print(f"file: {file_path}, input: {input_json}")
                f = open(file_path, 'r')
                data = f.read()
                f.close()
                template_data = Template(data)
                data = template_data.substitute(
                    {
                        'data_partition': data_partition,
                        'legal_tag': legal_tag,
                        'name': policy_name,
                        'domain': domain
                    })
                print(f"data: {data}")
                bdata = data.encode('utf-8')
                files = {'file': (input_json, bdata)}

                post_query(policy_id=policy_id, token=token, data_partition=data_partition, files=files, service_url=service_url, expect=expect)
                if opa_access:
                    post_query(policy_id=policy_id, token=token, data_partition=data_partition, files=files, expect=expect)

def get_legal_tags(token, data_partition):
    """
    get legal tags
    """

    url = conf.LEGAL_BASE_URL + "/api/legal/v1/legaltags?valid=true"
    headers = {'Authorization': 'Bearer ' + token,
        'data-partition-id': data_partition}
    r = requests.get(url, headers=headers)

    legal_tags = []

    for name in r.json()["legalTags"]:
        #print(name["name"])
        tags = name["name"]

        retval = common.get_legal_tag_info(data_partition_id=data_partition, access_token=token, legal_tags=[tags]) 
        if retval:
            #print(f"retval: {retval}")
            for retname in retval:
                x = retname["name"]
                assert x == tags, f"Expected to get {tags} from legal {retval}"
                legal_tags.append(tags)

                # we don't need all them, just enough to find issues
                if len(legal_tags) > 20:
                    return legal_tags

    return legal_tags
