# test_integration_035_translate.py
"""
To run the integration tests, start the OPA with sample policies first:

opa run --server --watch tests/integration/translate/policies/*.rego
"""

from override_depends import override_require_authorized_user, set_authorize_session, ADMIN_ONLY_SVC, USER_AND_ADMIN_SVC, USER_ONLY_SVC, OTHER_ONLY_SVC
import conf
from auth import auth
from main import app
import logging
import os
import signal
import stat
import subprocess
import sys
import time
from fastapi.testclient import TestClient
import pytest
import requests
import json
from pathlib import Path

# so the directly in which you run pytest or unittest doesn't matter as much
sys.path.append(os.path.abspath('..'))
sys.path.append('tests')

# local

# override dependency injection for authentication to entitlement service

TEST_DATA_DIR = Path(__file__).resolve().parent / 'templates'

client = TestClient(app)


@pytest.mark.dependency(name="require_token")
def test_require_token(token):
    assert token is not None, "No token provided on command line"


def opa_ast_to_es_request_body_perf(client):
    num_calls = 1000
    try:
        sys.stdout = open(os.devnull, "w")  # hide stdout
        start_seconds = time.perf_counter()
        for _ in range(0, num_calls):
            response = client.post(
                conf.SERVICE_BASE_PATH + "/translate",
                json={
                    "query": "data.search.allow == true",
                    "input": {
                        "operation": "view",
                        "groups": ["AAA", "BBB"],
                    },
                    "unknowns": ["input.record"],
                },
            )
            assert response.status_code == 200
        end_seconds = time.perf_counter()
        elapsed_seconds = end_seconds - start_seconds
        seconds_per_call = elapsed_seconds / num_calls
        print("seconds per call", seconds_per_call)
        target_seconds_per_call = 200 / 1000  # targeting 200 ms at the moment
        assert seconds_per_call < target_seconds_per_call
    finally:
        sys.stdout = sys.__stdout__  # restore stdout


def opa_ast_to_es_request_body(client, token, data_partition, data, service_url=None, status_code=200):
    url = conf.SERVICE_BASE_PATH + "/translate"
    if service_url:
        url = service_url + url
    logging.info(f"URL {url}")
    sdata = json.dumps(data)
    #bdata = sdata.encode('utf-8')
    #files = {'file': ("query.txt", bdata)}

    response = client.post(
        url,
        #files = files,
        data=sdata,
        headers={'Authorization': 'Bearer ' + token,
                 'data-partition-id': data_partition}
    )

    assert response.status_code == status_code, f"Expected {status_code} status code"
    return response.json()


@pytest.mark.dependency(depends=["require_token"])
def test_translate_local_and_service_url_search2(token, data_partition, service_url, opa_access):
    #pytest.skip("Skipping translate test")
    """ 
    This test requires search2.rego to be loaded previously in test_integration_020_put.py

    operation: view, create, update, delete, purge
    Send this:
    {
        "query": "data.osdu.partition[\"osdu\"].search2.allow == true",
        "input": {
            "operation": "view",
            "groups": ["AAA", "BBB"]
        },
        "unknowns": ["input.record"]
    }
    """
    data = {
        "query": "data.osdu.partition[\"" + data_partition + "\"].search2.allow == true",
        "input": {
            "operation": "view",
            "groups": ["AAA", "BBB"],
        },
        "unknowns": ["input.record"],
    }

    if opa_access:
        result = opa_ast_to_es_request_body(
            client=client, token=token, data_partition=data_partition, data=data)

        assert result == {
            "query": {
                "bool": {
                    "should": [
                        {"bool": {"filter": [
                            {"terms": {"acl.owners": ["AAA", "BBB"]}}]}},
                        {"bool": {"filter": [
                            {"terms": {"acl.viewers": ["AAA", "BBB"]}}]}},
                    ]
                }
            }
        }, f"Unexpected response from translate: {result}"

    result2 = opa_ast_to_es_request_body(client=requests,
                                         token=token,
                                         data_partition=data_partition,
                                         data=data,
                                         service_url=service_url)

    assert result2 == {
        "query": {
            "bool": {
                "should": [
                    {"bool": {"filter": [
                        {"terms": {"acl.owners": ["AAA", "BBB"]}}]}},
                    {"bool": {"filter": [
                        {"terms": {"acl.viewers": ["AAA", "BBB"]}}]}},
                ]
            }
        }
    }, f"Unexpected response from translate: {result}"

    if opa_access:
        assert result == result2, f"Unexpected response from translate: {result2}"


@pytest.mark.dependency(depends=["require_token"])
def test_translate_bad_operation(token, data_partition, service_url):
    """
    This test requires search2.rego to be loaded previously in test_integration_020_put.py
    """
    bad_operation = "badoperation"
    permitted_operations = ['view', 'create', 'update', 'delete', 'purge']
    data = {
        "query": "data.osdu.partition[\"" + data_partition + "\"].search2.allow == true",
        "input": {
            "operation": bad_operation,
            "groups": ["AAA", "BBB"],
        },
        "unknowns": ["input.record"],
    }
    result = opa_ast_to_es_request_body(client=requests,
                                        token=token,
                                        data_partition=data_partition,
                                        data=data,
                                        service_url=service_url,
                                        status_code=422)
    print(result)
    given = result['detail'][0]['ctx']['given']
    permitted = result['detail'][0]['ctx']['permitted']
    msg = result['detail'][0]['msg']
    assert "unexpected value" in msg, f"expected 'unexpected value' in msg of return json {msg}"
    assert bad_operation == given, f"expected 'bad operation' {bad_operation} in msg of return json"
    assert permitted == permitted_operations, f"expected list of permitted operations {permitted_operations} but got {permitted}"


@pytest.mark.dependency(depends=["require_token"])
def test_translate_not_permitted(token, data_partition, service_url):
    """
    This test requires checks for trying to access a partition that does not exist or you do not have access to
    """
    data = {
        "query": "data.osdu.partition[\"doesnotexist\"].search2.allow == true",
        "input": {
            "operation": "create",
            "groups": ["AAA", "BBB"],
        },
        "unknowns": ["input.record"],
    }
    result = opa_ast_to_es_request_body(client=requests,
                                        token=token,
                                        data_partition=data_partition,
                                        data=data,
                                        service_url=service_url,
                                        status_code=200)
    print(result)
    assert "query" in result, f"Expected query in {result}"
    assert "match_none" in str(result), f"Expected match_none in {result}"


@pytest.mark.dependency(depends=["require_token"])
def test_translate_does_not_exist(token, data_partition, service_url):
    """
    This test requires checks response on a policy that does not exist
    """
    data = {
        "query": "data.osdu.partition[\"" + data_partition + "\"].doesnotexist.allow == true",
        "input": {
            "operation": "view",
            "groups": ["AAA", "BBB"],
        },
        "unknowns": ["input.record"],
    }
    result = opa_ast_to_es_request_body(client=requests,
                                        token=token,
                                        data_partition=data_partition,
                                        data=data,
                                        service_url=service_url,
                                        status_code=404)
    print(result)
    assert "detail" in result, f"Expected detail in {result}"


@pytest.mark.dependency(depends=["require_token"])
def test_translate_no_groups(token, data_partition, service_url):
    """
    This test requires checks response on a policy that does not exist
    """
    data = {
        "query": "data.osdu.partition[\"" + data_partition + "\"].search2.allow == true",
        "input": {
            "operation": "view",
            "groups": [],
        },
        "unknowns": ["input.record"],
    }
    result = opa_ast_to_es_request_body(client=requests,
                                        token=token,
                                        data_partition=data_partition,
                                        data=data,
                                        service_url=service_url,
                                        status_code=200)
    print(result)
    assert "query" in result, f"Expected query in {result}"
    assert "match_none" in str(result), f"Expected match_none in {result}"
