from textwrap import indent
import pytest
import unittest
import responses
from fastapi.testclient import TestClient
import requests
import re
import logging
import sys
import os
import json
# so the directly in which you run pytest or unittest doesn't matter as much
sys.path.append(os.path.abspath('..'))
sys.path.append('tests')

# local
from main import app
from auth import auth
import conf
import testlib

#logging.basicConfig( stream=sys.stderr )
#logger = logging.getLogger("unittest" ).setLevel( logging.DEBUG )

# override dependency injection for authentication to entitlement service
from override_depends import override_require_authorized_user, set_authorize_session, ADMIN_ONLY_SVC, USER_AND_ADMIN_SVC, USER_ONLY_SVC, OTHER_ONLY_SVC

client = TestClient(app)

class TryUnitTesting(unittest.TestCase):

    #@responses.activate
    # def mock_entitlement(self):
    #     print("mock")
    #     entsvc = [{'name': 'service.policy.user', 'description': 'Policy user group', 'email': 'service.policy.user@osdu.group'}, {'name': 'service.policy.admin', 'description': 'Policy admin group', 'email': 'service.policy.admin@osdu.group'}]
    #     responses.add(
    #         responses.GET,
    #         conf.ENTITLEMENTS_GROUPS_API,
    #         json={"groups": entsvc},
    #         status=200,
    #     )

    def test_read_main_home_page(self):
        """Check Home Page"""
        response = client.get("/")
        self.assertEqual(response.status_code, 200, "Home Page is available")
        self.assertIn("Policy Service", response.text, "Expected policy service text on Home Page")

    def test_health_api_endpoint(self):
        """Check Health API"""
        response = client.get(conf.SERVICE_BASE_PATH+"/health")
        # Allow 503 for when OPA isn't available in unit testing
        assert response.status_code == 200 or response.status_code == 503
        self.assertIn("Healthy", response.text, "Expected Healthy in response text from /health")
    
    def test_docs_api_endpoint(self):
        """Check Docs endpoint"""
        response = client.get(conf.SERVICE_BASE_PATH+"/docs")
        self.assertEqual(response.status_code, 200, "Expected /docs to be available")

    def test_redocs_api_endpoint(self):
        """Check ReDocs endpoint"""
        response = client.get(conf.SERVICE_BASE_PATH+"/redoc")
        self.assertEqual(response.status_code, 200, "Expected Redoc to be available")

    def test_openapi_json_endpoint(self):
        """Check openapi json is available"""
        response = client.get(conf.SERVICE_BASE_PATH+"/openapi.json")
        self.assertEqual(response.status_code, 200, "Expected OpenAPI.json to be available") 

    def requires_auth(self, method, URL):
        """ Should always return 403 """
        with self.subTest(URL=URL):
            response = method(URL)
            self.assertEqual(response.status_code, 403, f"Not sending token to our API should result in 403 Forbidden {URL}")

    def test_api_endpoints_make_sure_they_require_auth(self):
        self.requires_auth(client.get, conf.SERVICE_BASE_PATH+"/policies")
        self.requires_auth(client.get, conf.SERVICE_BASE_PATH+"/policies/xyz")
        #self.requires_auth(client.put, conf.SERVICE_BASE_PATH+"/policies/xyz")
        #telf.requires_auth(client.delete, conf.SERVICE_BASE_PATH+"/policies/xyz")
        self.requires_auth(client.get, conf.SERVICE_BASE_PATH+"/policies/osdu/instance/xyz")
        self.requires_auth(client.get, conf.SERVICE_BASE_PATH+"/policies/osdu/partition/xyz/xyz")
        self.requires_auth(client.put, conf.SERVICE_BASE_PATH+"/policies/osdu/partition/xyz/xyz")
        self.requires_auth(client.delete, conf.SERVICE_BASE_PATH+"/policies/osdu/partition/xyz/xyz")
        self.requires_auth(client.post, conf.SERVICE_BASE_PATH+"/evaluations/query")
        self.requires_auth(client.post, conf.SERVICE_BASE_PATH+"/translate")

    @unittest.skipIf(conf.ENABLE_DEV_DIAGNOSTICS == False, "Skipping diagnostic test - ENABLE_DEV_DIAGNOSTICS")
    @pytest.mark.opa_access
    @set_authorize_session(ADMIN_ONLY_SVC)
    def test_local_config_as_admin_only(self):
        """Check config API as a member of only admin group"""
        r = client.get("/diag/config")
        print(f"config text: {r.text}")
        print(f"text: {r.text}",file=sys.stderr)
        self.assertEqual(r.status_code, 200, "Config Page is available")

    @pytest.mark.opa_access
    @unittest.skipIf(conf.ENABLE_DEV_DIAGNOSTICS == False, "Skipping diagnostic test - ENABLE_DEV_DIAGNOSTICS")
    @set_authorize_session(USER_AND_ADMIN_SVC)
    def test_config_as_user_and_admin(self):
        """Check config API as a member of both user and admin groups"""
        r = client.get("/diag/config")
        self.assertEqual(r.status_code, 200, "Expected config Page is available")
        if r.ok:
            js = json.dumps(r.json(), indent=4)
            print(f"{js}")

    @unittest.skipIf(conf.ENABLE_DEV_DIAGNOSTICS == False, "Skipping diagnostic test - ENABLE_DEV_DIAGNOSTICS")
    @set_authorize_session(USER_ONLY_SVC)
    def test_config_as_user_only(self):
        """Check config API as member of user group"""
        r = client.get("/diag/config")
        print(f"text: {r.text}")
        self.assertEqual(r.status_code, 403, "Expected config Page is available but forbidden")

    @pytest.mark.opa_access
    @unittest.skipIf(conf.ENABLE_DEV_DIAGNOSTICS == False, "Skipping diagnostic test - ENABLE_DEV_DIAGNOSTICS")
    @set_authorize_session(OTHER_ONLY_SVC)
    def test_config_as_other(self):
        """Check config API as not a member of user and admin groups"""
        r = client.get("/diag/config")
        self.assertEqual(r.status_code, 403, "Expected config is available but forbidden")
        print(f"text: {r.text}")

    @unittest.skipIf(conf.ENABLE_DEV_DIAGNOSTICS == False, "Skipping diagnostic test - ENABLE_DEV_DIAGNOSTICS")
    @set_authorize_session()
    def test_config_as_none(self):
        """Check config API as not a member of any group"""
        r = client.get("/diag/config")
        print(f"text: {r.text}")
        self.assertEqual(r.status_code, 403, "Expected config is available but forbidden")
    
    @set_authorize_session(USER_ONLY_SVC)
    def test_info(self):
        """Check info API endpoint"""
        r = client.get(conf.SERVICE_BASE_PATH + "/info")
        print(f"text: {r.text}")
        self.assertEqual(r.status_code, 200, "Expected /info is available")

    @set_authorize_session(USER_ONLY_SVC)
    def test_mock_read_policies(self):
        """Check /policies API endpoint"""
        os.environ['CLOUD_PROVIDER'] = conf.MOCK
        r = client.get(conf.SERVICE_BASE_PATH + "/policies")
        print(f"text: {r.text}")
        self.assertEqual(r.status_code, 200, "Expected /policies is available")

    # at some point this should be expanded to test permissions
    @set_authorize_session(OTHER_ONLY_SVC, "other_partition_id")
    def test_mock_read_a_policy(self):
        """Check /policies API endpoint"""
        os.environ['CLOUD_PROVIDER'] = conf.MOCK
        r = client.get(conf.SERVICE_BASE_PATH + "/policies/xyz")
        print(f"text: {r.text}")
        self.assertEqual(r.status_code, 404, "Not expected to be found")

    @set_authorize_session(ADMIN_ONLY_SVC, "osdu")
    def test_mock_put_all_policies_test_data(self):
        os.environ['CLOUD_PROVIDER'] = conf.MOCK
        testlib.put_policies_test_data(client=client, token="MOCK", data_partition="osdu")

if __name__ == "__main__":
    logging.basicConfig( stream=sys.stderr )
    logging.getLogger("unittest" ).setLevel( logging.DEBUG )
    unittest.main()
