# Enforcing OPA policies in another context

## Introduction

While OPA is able to perform evaluation solely based on (pre)defined policies and user input, certain evaluations require additional "context" from the caller to complete. In particular, this applies to search operations, where the objective is to enumerate all documents that satisfy a particular policy.

The straightforward implementation in this scenario would be to retrieve all documents from e.g. the database, and then pass this context to OPA as part of the user input. However, this is not always feasible, as the number of documents can be large. Furthermore, it is clearly inefficient if the caller is interested in the first few entries, as is the case with pagination.

The ideal solution would be to somehow "map" the OPA policies to a form that can be evaluated by the database itself -- ElasticSearch queries for example. OPA addresses this use case via the [Compile API](https://www.openpolicyagent.org/docs/latest/rest-api/#compile-api). By providing the query, user input, and the unknowns, OPA can return the AST of a simplified version of the policy. The Policy Service takes this one step further and translates the AST into the native database query. The caller is then responsible to take the translated query and execute it against the data source.

Currently, only translation to ElasticSearch queries is supported. The feature should be considered experimental. Other targets, such as SQL, might be supported in a future release.

## Example

In the default configuration, the API is exposed at `/api/policy/v1/translate`. As an example, assume that there is a predefined OPA policy `search.rego`, and it is desired to have all ElasticSearch queries retrieve only documents that pass this policy.

```
package search

allow {
  input.record.acl.viewers[_] == input.user.groups[_]
}
```

This policy says that, the `data.search.allow` rule will return true if any one of the members in the `acl.viewers` array equals any one of the members in `user.groups` array. While `input` is a designated global variable in OPA that refers to the input, the `record` and `user` variables are arbitrary and is up to the caller to define in the policy.

The caller can then POST a request body:

```
{
  "query": "data.search.allow == true",
  "input": {
    "user": {
      "groups": ["sample-group-1"],
    },
  },
  "unknowns": ["input.record"]
}
```
This request body asks OPA the following question: Given the user profile containing group membership information, what query should be performed to fill up the variable `input.record`, such that the policy `data.search.allow` evaluates to true?

Internally, the Policy Service forwards the information to OPA for the partial evaluation process and gets back an AST. The Policy Service then translates the AST into ElasticSearch queries and returns it to the caller. The output would look something like this:

```
{
  "query": {
    "bool": {
      "should": [{
        "bool": {
          "filter": [{
            "term": {
              "acl.viewers": "sample-group-1"
            }
          }]
        }
      }]
    }
  }
}
```

It basically says to filter and return only documents that have `sample-group-1` in the `acl.viewers` array (it is assumed that the field has a keyword type family). The caller can then take this query, append it to the original ElasticSearch query, and finally evaluate it against ElasticSearch.

## Note:
Translate is known to work with basic policies. Complex policies, including ones using deny may have issues. It may not work for all scenarios.
