#!/usr/bin/env python3.9
# Policy Service Admin CLI
from time import time
import typer
from typing import List, Optional
from types import SimpleNamespace
#from rich import print
from rich.table import Table
from rich.console import Console
from rich.tree import Tree
from rich.panel import Panel
import rich.progress
from enum import Enum
import requests
from pathlib import Path
import os
import sys
import re
import json
import getpass
import random
from string import Template
import difflib
from uuid_extensions import uuid7str
from rego import ast, walk
import opa
import search_cli

__version__ = '0.0.4'
__app_name__ = 'OSDU Policy Service AdminCLI'

cli = typer.Typer(rich_markup_mode="rich", help=__app_name__)
console = Console()
error_console = Console(stderr=True, style="bold red")

# setup simple output class
class OutputType(str, Enum):
    simple = "simple"
    fancy = "fancy"
    tree = "tree"
    excel = "excel"

def _version_callback(value: bool) -> None:
    if value:
        typer.echo(f"{__app_name__} v{__version__}")
        raise typer.Exit()

def headers(ctx: typer.Context, content_type_json=False):
    """
    build headers for OSDU request
    """
    if content_type_json:
        headers={'Authorization': 'Bearer ' + ctx.obj.token,
            'data-partition-id': ctx.obj.data_partition,
            'accept': 'application/json',
            'content-type': 'application/json'
            }
    else:
        headers={'Authorization': 'Bearer ' + ctx.obj.token,
            'data-partition-id': ctx.obj.data_partition,
            'accept': 'application/json',
            }
    if ctx.obj.uuid:
        headers['Correlation-id'] = ctx.obj.uuid
    if ctx.obj.x_user_id:
        headers['x-user-id'] = ctx.obj.x_user_id
    if ctx.obj.user_id:
        headers['x-local-user-id'] = ctx.obj.user_id
    if ctx.obj.debug:
        console.print(f"headers: {headers}")
    return headers

def request_groups(ctx: typer.Context):
    """
    return list of groups from entitlements service
    """
    try:
        response = requests.get(ctx.obj.entitlements_url + "/groups", headers = headers(ctx))
    except requests.exceptions.RequestException as e: 
        raise SystemExit(e)

    if not response.ok:
        if "message" in response.text:
            message = json.loads(response.text)["message"]
            error_console.print(f"Error: {message}")
        else:
            error_console.print(f"Error: An error occurred when talking to {ctx.obj.entitlements_url}")
        return ""
    jdata = response.json()
    if "groups" in jdata:
        return jdata['groups']

def get_policies(ctx: typer.Context):
    """
    Handle request to /policies
    """
    try:
        r = requests.get(ctx.obj.url + "/policies",
            timeout=10,
            headers=headers(ctx))
    except requests.exceptions.HTTPError:
        error_console.print(f"Error: endpoint {ctx.obj.url}: HTTPError")
        return
    except requests.exceptions.ConnectionError:
        error_console.print(f"Error: endpoint {ctx.obj.url}: ConnectionError")
        return
    except requests.exceptions.Timeout:
        error_console.print(f"Error: endpoint {ctx.obj.url}: Timeout")
        return
    except requests.exceptions.RequestException:
        error_console.print(f"Error: endpoint {ctx.obj.url}: RequestException")
        return

    if "result" in r.text and r.ok:
        return(r.json()["result"])
    else:
        if "detail" in r.text:
            error_console.print(f"Error: {r.json()['detail']}")
        else:
            error_console.print(f"Error: {r.text}")
        raise typer.Exit(1) # non-zero exit status

def delete_partition_policy(ctx: typer.Context, policy_id: str):
    """
    handle request to delete a policy from a partition
    """
    try:
        r = requests.delete(ctx.obj.url + "/policies/osdu/partition/" + ctx.obj.data_partition + '/' + policy_id,
            headers=headers(ctx))
    except requests.exceptions.RequestException as e: 
        raise SystemExit(e)

    if ctx.obj.debug:
        console.print(f"debug: {r.json()}")
    if r.ok:
        if "result" in r.text:
            return True
    else:
        if "detail" in r.text:
            error_console.print(f"Error: {r.json()['detail']}")
        else:
            error_console.print(f"Error: {r.text}")
        raise typer.Exit(1) # non-zero exit status

def add_partition_policy(ctx: typer.Context, policy_id: str, files: dict):
    """
    handle request to add a policy to a partition
    """
    try:
        r = requests.put(ctx.obj.url + "/policies/osdu/partition/" + ctx.obj.data_partition + '/' + policy_id,
            files=files,
            headers=headers(ctx))
    except requests.exceptions.RequestException as e: 
        raise SystemExit(e)

    if r.ok:
        return r.text
    else:
        if "detail" in r.text:
            error_console.print(f"Error: {r.json()['detail']}")
        else:
            error_console.print(f"Error: {r.text}")
        raise typer.Exit(1) # non-zero exit status

def get_partition_policy(ctx: typer.Context, policy_id: str, quiet=False):
    """
    handle request to get partition policy
    """
    try:
        r = requests.get(ctx.obj.url + "/policies/osdu/partition/" + ctx.obj.data_partition + '/' + policy_id,
            headers=headers(ctx))
    except requests.exceptions.RequestException as e: 
        raise SystemExit(e)

    if r.ok:
        if "result" in r.text:
            return(r.json()["result"])
    else:
        if not quiet and "detail" in r.text:
            error_console.print(f"Error: {r.json()['detail']}")
        else:
            error_console.print(f"Error: {r.text}")
        #raise typer.Exit(1) # non-zero exit status

def get_instance_policy(ctx: typer.Context, policy_id: str, quiet=False):
    """
    handle request to get instance policy
    """
    try:
        r = requests.get(ctx.obj.url + "/policies/osdu/instance/" + policy_id,
            headers=headers(ctx))
    except requests.exceptions.RequestException as e: 
        raise SystemExit(e)

    if r.ok:
        if "result" in r.text:
            return(r.json()["result"])
    else:
        if not quiet and "detail" in r.text:
            error_console.print(f"Error: {r.json()['detail']}")
        else:
            error_console.print(f"Error: {r.text}")
        raise typer.Exit(1) # non-zero exit status

def evaluations_query(ctx: typer.Context, policy_id: str, files: dict):
    """
    handle request to eval query
    """
    params = {
        'policy_id': policy_id,
        'include_auth': True
        }

    try:
        r = requests.post(ctx.obj.url + "/evaluations/query",
            params=params,
            files=files,
            headers=headers(ctx))
    except requests.exceptions.RequestException as e: 
        raise SystemExit(e)

    if r.ok:
        if "result" in r.text:
            return(r.json()["result"])
    else:
        if "detail" in r.text:
            error_console.print(f"Error: {r.json()['detail']}")
        else:
            error_console.print(f"Error: {r.text}")
        raise typer.Exit(r.status_code) # non-zero exit status

@cli.command(rich_help_panel="Utils")
def groups(ctx: typer.Context,
        all: bool = False,
        domain: bool = False,
        domain_suffix: bool = False):
    """
    Show groups of current auth context.
    """
    retgroups = request_groups(ctx)
    if all:
        console.print(retgroups)
    elif domain:
        domain=retgroups[0]["email"].split('@')[1]
        domain_ending = domain.removeprefix(ctx.obj.data_partition + ".")
        console.print(f"fulldomain: {domain}\ndomain: {domain_ending}")
    elif domain_suffix:
        domain=retgroups[0]["email"].split('@')[1]
        domain_ending = domain.removeprefix(ctx.obj.data_partition + ".")
        console.print(domain_ending)
    else:
        for grp in retgroups:
            if "policy" in str(grp):
                console.print(grp)

def get_domain_suffix(ctx):
    """
    get domain
    """
    retgroups = request_groups(ctx)
    domain=retgroups[0]["email"].split('@')[1]
    domain_ending = domain.removeprefix(ctx.obj.data_partition + ".")
    return domain_ending

def is_json(myjson):
    """
    Is this string valid json - detect problems before sending it to policy service
    """
    try:
        json.loads(myjson)
    except ValueError as e:
        return False
    return True

@cli.command(rich_help_panel="Policy Developer Utils")
def translate(ctx: typer.Context,
        policy_id: str= typer.Argument(...),
        output: OutputType = typer.Option("fancy", help="Output style"),
        file: Path = typer.Option(...,
            "-f", "--file",
            exists=True,
            file_okay=True,
            dir_okay=False,
            writable=False,
            readable=True,
            resolve_path=True,
            help="json file"
        ),
        force: bool = typer.Option(False, "--force", help="No confirmation prompt"),
        template: bool = typer.Option(False, "--template", "-t", help="Input is a template to be rendered")
):
    """
    Translate testing utility.

    Requires valid json file.

        If using option --template:

        ${data_partition} = data_partition associated with your token
        ${name} = short name you give as a policy_id on the command line

    Output fancy (default) draws panels, simple just displays text
    """
    with rich.progress.open(file, "r") as f:
        data = f.read()
    f.close()
    short_policy_id = os.path.basename(policy_id)
    if template:
        template_data = Template(data)
        data = template_data.substitute(
            {'data_partition': ctx.obj.data_partition,
            'name': short_policy_id.removesuffix('.rego')
        })
        if output == output.fancy:
            console.print(Panel(data, title="rendered data", highlight=True))
        else:
            print(data)
        if not force:
            typer.confirm(f"translate above data against {short_policy_id} in {ctx.obj.data_partition}", abort=True)

    if not is_json(data):
        error_console.print("Error: Not valid json")
        raise typer.Exit(1)
    
    try:
        with console.status("Translating..."):
            r = requests.post(ctx.obj.url + "/translate",
                data=data,
                headers=headers(ctx))
    except requests.exceptions.RequestException as e: 
        error_console.print(f"Error: {e}")
        raise SystemExit(e)

    if r.ok:
        if output == output.fancy:
            console.print(r.json())
        else:
            print(r.json())
    else:
        if "detail" in r.text:
            error_console.print(f"Error: {r.json()['detail']} {r.status_code}")
        else:
            error_console.print(f"Error: {r.text}")
        raise typer.Exit(1) # non-zero exit status

@cli.command(rich_help_panel="Policy Commands")
def eval(ctx: typer.Context,
        policy_id: str= typer.Argument(None),
        file: Path = typer.Option(...,
            "-f", "--file",
            exists=True,
            file_okay=True,
            dir_okay=False,
            writable=False,
            readable=True,
            resolve_path=True,
        ),
        template: bool = typer.Option(False, "--template", "-t", help="Input is a template to be rendered"),
        output: OutputType = typer.Option("fancy", help="Output style"),
        force: bool = typer.Option(False, "--force", help="No confirmation prompt"),
        legal_tag: str = None,
        domain: str = None
        ):
    """
    Evaluate Policy.

        If using option --template:

        ${data_partition} = data_partition associated with your token
        ${name} = short name you give as a policy_id on the command line
        ${legal_tag} = value of --legal_tag on the command line
        ${domain} = value of --domain the command line, if none provide admincli will determine

        Output fancy (default) draws panels, simple just displays text

        This command will allow the policy service to add/replace any auth details in json using the include_auth capability in API.
    """
    if policy_id.startswith('osdu/instance/'):
       error_console.print("Error: Instance policies not supported")
       raise typer.Abort()

    if domain is None:
        domain = get_domain_suffix(ctx)
        #console.print(f"domain is {domain}")

    short_policy_id = os.path.basename(policy_id)

    if not short_policy_id.endswith(".rego"):
        short_policy_id = short_policy_id + ".rego"

    with rich.progress.open(file, "r") as f:
        data = f.read()
    f.close()

    if template:
        template_data = Template(data)
        data = template_data.substitute(
            {
                'data_partition': ctx.obj.data_partition,
                'name': short_policy_id.removesuffix('.rego'),
                'legal_tag': legal_tag,
                'domain': domain
            })
        if output == output.fancy:
            console.print(Panel(data, title="rendered policy", highlight=True))
        else:
            print(data)

    if not is_json(data):
        error_console.print("Error: Not valid json")
        raise typer.Exit(1)

    if not force:
        typer.confirm(f"evaluate against {short_policy_id} in {ctx.obj.data_partition}", abort=True)

    bdata = data.encode('utf-8')
    files = {'file': (short_policy_id, bdata)}
    eval_result = evaluations_query(ctx = ctx, policy_id=policy_id, files=files)
    console.print(eval_result)

def search_policies_full(result: list, search: str):
    """
    search helper
    """
    retlist = []
    for pol in result:
        if re.search(search, str(pol)):
            retlist.append({'id':pol['id'], 'raw': pol['raw']})
    return retlist

def search_policies_idonly(result: list, search: str):
    """
    search helper in policy name only
    """
    retlist = []
    for pol in result:
        if re.search(search, pol['id']):
            retlist.append({'id':pol['id'], 'raw': pol['raw']})
    return retlist
    
def display_all_policies(ctx: typer.Context,
    output: OutputType = OutputType.fancy,
    raw: bool = False,
    search_name = None,
    search = None
    ):
    """
    display policies
    """
    
    # retrieve all policy ids
    with console.status("Getting policies..."):
        result = get_policies(ctx)

    if search_name:
       with console.status("Searching policies by name..."):
            result = search_policies_idonly(result, search_name)

    if search:
       with console.status("Searching policies..."):
            result = search_policies_full(result, search)

    if result:
        if raw:
            if output == output.simple:
                print(result)
            else:
                console.print(result)
            raise typer.Exit()

        # tree output
        if output == output.tree:
            tree = Tree("Policy Tree")
            partition_tree = tree.add("osdu/partition/")
            instance_tree = tree.add("osdu/instance/")
            data_partition_tree = partition_tree.add(ctx.obj.data_partition)
            for pol in result:
                #pol['id'].basename
                short_policy_id = os.path.basename(pol['id'])
                dir = os.path.dirname(pol['id'])
                if pol['id'].startswith("osdu/instance/"):
                    instance_tree.add(short_policy_id)
                else:
                    data_partition_tree.add(short_policy_id)
            console.print(tree)
            raise typer.Exit()

        try: 
            # silly math for figuring out terminal size
            width, height = os.get_terminal_size()
        except:
            # not a tty... no worries just simple output
            output = output.simple

        # simple output
        if output == output.simple:
            for pol in result:
                print(pol['id'])
            raise typer.Exit()
        
        # else let's make it look pretty
        if search or search_name:
            table = Table(title="Search results")
        else:
            table = Table(title="Lookup of all policy IDs")
        name_width=0

        for pol in result:
            id_width = len(pol['id'])
            if id_width > name_width:
                name_width = id_width

        # silly math for figuring out terminal size
        preview_width = width - 6 - name_width
        table.add_column("Name", justify="full", style="cyan", no_wrap=True, min_width=name_width)
        table.add_column("Preview", justify="center", style="green", no_wrap=True, max_width=preview_width)

        for pol in result:
            preview=pol['raw'].replace("\n", " ").strip()
            preview=" ".join(preview.split())
            if preview.startswith("p"):
                table.add_row(pol['id'], preview.strip())
            else:
                console.print("Error: unexpected preview data")
        console.print(table)

@cli.command("ls", rich_help_panel="Policy Commands")
def list(ctx: typer.Context,
        policy_list: List[str]= typer.Argument(None),
        raw: bool = typer.Option(False, "--raw", help="Full output from policy service"),
        quiet: bool = typer.Option(False, "--quiet", help="Don't display policy"),
        output: OutputType = typer.Option("fancy", help="Output style"),
        search: str = typer.Option("", "--search", "-s", help="Regex search string", rich_help_panel="Search Options"),
        search_name: str = typer.Option("", "--name", "-n", help="Regex search string in policy name only", rich_help_panel="Search Options"),
        download: bool = typer.Option(False, "--download", "-d", help="download policy to file")
        ):
    '''
    Lookup policies or a policy.

    If policy_list is - on command-line, the list will be read from standard-in
    '''
    if output == output.excel:
        error_console.print("Error: Not supported output type")
        raise typer.Exit(1) # non-zero exit status

    if policy_list == ['-']:
        policy_list = sys.stdin.read().strip().split()

    if not policy_list:
        display_all_policies(ctx = ctx, output = output, raw = raw, search=search, search_name=search_name)
    else:
        for policy_id in policy_list:
            if '*' in policy_id:
                display_all_policies(ctx = ctx, output = output, raw = raw, search=policy_id)
                return
            # if includes a name without .rego, let's add it to make it easier to use cli
            if not policy_id.endswith(".rego"):
                policy_id = policy_id + ".rego"

            short_policy_id = os.path.basename(policy_id)

            result = lookup_policy(ctx=ctx, policy_id=policy_id, quiet=quiet)
            if result:
                if raw:
                    if download:
                        filename = short_policy_id.removesuffix('rego') + 'json'
                        with console.status(f"Saving {filename}..."):
                            f = open(filename, "w")
                            f.write(str(result))
                            f.close()
                        if not quiet:
                            console.print(f"{policy_id} raw saved as '{filename}'")
                    elif quiet:
                        pass
                    elif output == output.fancy:
                        console.print(result)
                    else: # simple
                        print(result)
                else: 
                    if download:
                        with console.status(f"Saving {short_policy_id}..."):
                            f = open(short_policy_id, "w")
                            f.write(str(result['raw']))
                            f.close()
                        if not quiet:
                            console.print(f"{policy_id} raw saved as '{short_policy_id}'")
                    elif quiet:
                        pass
                    elif output == output.fancy:
                        console.print(result['raw'])
                    else: # simple
                        print(result['raw'])
            else:
                raise typer.Exit(1) # non-zero exit status

@cli.command(hidden=True, rich_help_panel="Policy Developer Utils")
def test(ctx: typer.Context, policy_id: str = typer.Option("", '-p')):
    """
    Do some tests. Hidden command
    """
    console.print(ctx.obj)

@cli.command(rich_help_panel="Policy Commands")
def health(ctx: typer.Context):
    """
    Is Policy service healthy
    """
    try:
        with console.status("Getting health..."):
            r = requests.get(ctx.obj.url + "/health", headers=headers(ctx))
    except requests.exceptions.RequestException as e: 
        raise SystemExit(e)

    if r.ok:
        console.print(":thumbsup-emoji:",r.text)
    else:
        error_console.print(r.text)
        raise typer.Exit(1) # non-zero exit status

@cli.command(rich_help_panel="Utils")
def info(ctx: typer.Context,
    service: str = typer.Option("policy", "-s", "--service", help="Get search info")
    ):
    """ 
    Info on Dataset, Entitlement, Legal, Policy Service, Search, and Storage
    """
    if "storage" in service:
        get_info(ctx=ctx, base_url=ctx.obj.storage_url)
    elif "search" in service:
        get_info(ctx=ctx, base_url=ctx.obj.search_url)
    elif "entitlement" in service:
        get_info(ctx=ctx, base_url=ctx.obj.entitlements_url)
    elif "legal" in service:
        get_info(ctx=ctx, base_url=ctx.obj.legal_url)
    elif "policy" in service:
        get_info(ctx=ctx, base_url=ctx.obj.url)
    elif "dataset" in service:
        get_info(ctx=ctx, base_url=ctx.obj.dataset_url)
    else:
        error_console.print(f"Error: Unsupported service {service}")
        raise typer.Exit(1) # non-zero exit status

@cli.command(rich_help_panel="Policy Developer Utils")
def diff(ctx: typer.Context,
    policy_id_1: str = typer.Argument(...),
    policy_id_2: str = typer.Argument(...),
    n: int = typer.Option(3, "-n", help="Number of adjacent lines to show"),
    ):
    """
    Compare two policies, show the delta in a context diff format.

    Context diffs are a compact way of showing just the lines that have changed plus
    a few lines of context. The changes are shown in a before/after style.
    The number of context lines is set by n which defaults to three.
    """

    # if without .rego, let's add it to make it easier to use cli
    if not policy_id_1.endswith(".rego"):
        policy_id_1 = policy_id_1 + ".rego"

    if not policy_id_2.endswith(".rego"):
        policy_id_2 = policy_id_2 + ".rego"

    result1 = lookup_policy(ctx, policy_id_1)
    result2 = lookup_policy(ctx, policy_id_2)

    differences = difflib.context_diff(
        a=result1['raw'].splitlines(),
        b=result2['raw'].splitlines(),
        fromfile=policy_id_1,
        tofile=policy_id_2,
        n=n)
    console.print('\n'.join(differences))

def lookup_policy(ctx: typer.Context, policy_id, quiet=False):
    short_policy_id = os.path.basename(policy_id)
    if "osdu/instance/" in policy_id: # lookup instance policy
        with console.status(f"Getting instance policy {short_policy_id}..."):
            result = get_instance_policy(ctx, short_policy_id, quiet)
    elif "osdu/partition/" in policy_id: # lookup partition policy
        with console.status(f"Getting partition policy {short_policy_id}..."):
            result = get_partition_policy(ctx, short_policy_id, quiet)
    else: # assume it's a partition policy
        with console.status(f"Getting partition policy {short_policy_id}..."):
            result = get_partition_policy(ctx, short_policy_id, quiet)
        if not result:
            # nope report the error and let's still check for a instance policy
            if not quiet:
                error_console.print(f"[green]Checking for '{short_policy_id}' in instance policies[/]...")
            result = get_instance_policy(ctx, short_policy_id, quiet)
    if not result:
        if not quiet:
            error_console.print(f"Error: Unable to find {policy_id}")
        raise typer.Exit(1) # non-zero exit status

    return result

@cli.command(rich_help_panel="Policy Developer Utils")
def compile(ctx: typer.Context,
    file: Path = typer.Option(...,
        "-f", "--file",
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
        ),
    metrics: bool = False,
    opt_walk: bool = typer.Option(False, "--walk"),
    instrument: bool = False,
    template: bool = typer.Option(False, "--template", "-t", help="Input is a template to be rendered"),
    force: bool = typer.Option(False, "--force", help="No confirmation prompt")
    ):
    """
    Use OPA's Compile API to partially evaluate a query.

    Diagnostic compile on Policy Service. Requires service have ENABLE_DEV_DIAGNOSTICS enabled.
    """
    with rich.progress.open(file, "r") as f:
        data = f.read()
    f.close()
    if template:
        template_data = Template(data)
        data = template_data.substitute(
            {'data_partition': ctx.obj.data_partition}
        )
        console.print(Panel(data, title="rendered policy", highlight=True))
    if not force:
        typer.confirm(f"compile in {ctx.obj.data_partition}", abort=True)

    bdata = data.encode('utf-8')
    files = {'file': ("compile.json", bdata)}
    params = {'metrics': metrics, 'instrument': instrument}

    try:
        r = requests.post(ctx.obj.diag_url + "/compile", files=files, params=params, headers=headers(ctx))
    except requests.exceptions.RequestException as e: 
        raise SystemExit(e)

    if r.ok:
        console.print(r.json())
    else:
        error_console.print(r.text, r.status_code)
        raise typer.Exit(1) # non-zero exit status

    if opt_walk:
        # Load the resulting set of query ASTs out of the JSON response.
        if "result" in r.json() and "queries" in r.json()["result"]:
            tree = r.json()["result"]["queries"]
            console.print(Panel(str(tree), title="Abstract Syntax Tree", highlight=True))
            qs = ast.QuerySet.from_data(tree)
            # Pretty print the ASTs.
            walk.pretty_print(qs)
        else:
            error_console.print("Error: unexpected response from policy service. Check if policy exists")
            raise typer.Exit(1) # non-zero exit status

@cli.command(rich_help_panel="Policy Developer Utils")
def config(ctx: typer.Context):
    """
    Diagnostic config on Policy Service.
    
    Requires service have ENABLE_DEV_DIAGNOSTICS enabled.
    """
    try:
        r = requests.get(ctx.obj.diag_url + "/config", headers=headers(ctx))
    except requests.exceptions.RequestException as e: 
        raise SystemExit(e)

    if r.ok:
        console.print(r.json())
    else:
        error_console.print(r.text)
        raise typer.Exit(1) # non-zero exit status

def get_legal_tags(ctx: typer.Context, tag_name: str = None):
    """
    Get legal tags json
    """
    try:
        if tag_name:
            with console.status(f"Getting legal tag {tag_name}..."):
                r = requests.get(ctx.obj.legal_url + "/legaltags/" + tag_name, headers = headers(ctx))
        else:
            with console.status("Getting legaltags..."):
                r = requests.get(ctx.obj.legal_url + "/legaltags" + "?valid=true", headers = headers(ctx))
    except requests.exceptions.RequestException as e: 
        raise SystemExit(e)
    
    if r.ok:
        return r.json()
    else:
        error_console.print("Error:", r.text)
        raise typer.Exit(1) # non-zero exit status

def create_legal_tag(ctx: typer.Context, json_data, update=False):
    """
    Create or update a legal tag
    """
    try:
        if update:
            with console.status(f"Updating legal tag..."):
                r = requests.put(ctx.obj.legal_url + "/legaltags", json=json.loads(json_data), headers = headers(ctx, content_type_json=True))
        else:
            with console.status(f"Creating legal tag..."):
                r = requests.post(ctx.obj.legal_url + "/legaltags", json=json.loads(json_data), headers = headers(ctx, content_type_json=True))
    except requests.exceptions.RequestException as e: 
        raise SystemExit(e)
    
    if r.ok:
        return r.json()
    else:
        if r.status_code == 409:
            error_console.print("Error: Legal tag already exists")
            raise typer.Exit(2) # non-zero exit status
        else:
            error_console.print("Error:", r.text, r.status_code)
            raise typer.Exit(1) # non-zero exit status

def get_a_random_legal_tag(ctx: typer.Context):
    """
    Get a random legal tag
    """
    r_json = get_legal_tags(ctx=ctx)
    lucky_tag = random.choice(r_json["legalTags"])
    return lucky_tag["name"].strip()

@cli.command(rich_help_panel="Utils")
def add_legal_tag(ctx: typer.Context,
    file: Path = typer.Option(...,
        "-f", "--file",
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
        ),
    update: bool = typer.Option(False, "--update", help="Update existing legal tag")
    ):
    """
    Create or Update a legal tag

    Updating legal tags: Currently the legaltags API only allows the updating of the description, contract ID, expiration date and extensionProperties properties
    """

    with rich.progress.open(file, "r") as f:
        data = f.read()
    f.close()

    if not is_json(data):
        error_console.print("Error: Not valid json")
        raise typer.Exit(1)

    r_json = create_legal_tag(ctx=ctx, json_data=data, update=update)
    console.print(json.dumps(r_json, indent=4))

@cli.command(rich_help_panel="Utils")
def legal_tags(ctx: typer.Context,
        tag_list: List[str]= typer.Argument(None, help="List of tags to show"),
        raw: bool = typer.Option(False, "--raw", help="Get raw json output"),
        random: bool = typer.Option(False, "--random", help="Get a random legal tag"),
        limit: int = typer.Option(None, help="Limit number")):
    """
    Show legal tags.
    """
    
    if tag_list == ['-']:
        tag_list = sys.stdin.read().strip().split()

    if tag_list:
        for tag in tag_list:
            r_json = get_legal_tags(ctx=ctx, tag_name=tag)
            console.print(json.dumps(r_json, indent=4))
    else:
        if raw:
            r_json = get_legal_tags(ctx=ctx)
            console.print(json.dumps(r_json, indent=4))
        elif random:
            console.print(get_a_random_legal_tag(ctx=ctx))
        else:
            r_json = get_legal_tags(ctx=ctx)
            count = 0
            print(type(r_json["legalTags"]))
            for name in r_json["legalTags"]:
                console.print(name["name"])
                count = count + 1
                if count == limit:
                    break

@cli.command("opa-add", rich_help_panel="Policy Developer Utils")
def add_to_opa(ctx: typer.Context,
    policy_id: str = typer.Argument(...),
    file: Path = typer.Option(...,
        "-f", "--file",
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
        ),
    template: bool = typer.Option(False, "--template", "-t", help="Input is a template to be rendered"),
    force: bool = typer.Option(False, "--force", help="No confirmation prompt"),
    url: str = typer.Option("http://localhost:8181", help="Base URL to connect to OPA")
    ):
    """
    [green]Add or update[/green] a policy directly in OPA :sparkles: for LOCAL testing/development
    """
    if ctx.obj.debug:
        console.print(f"url: {url}")
    
    if policy_id.startswith('osdu/instance/'):
        policy_type = "instance"
    elif policy_id.startswith('osdu/partition/'):
        policy_type = "partition"
    else:
        error_console.print("Error: policy_id must start with osdu/instance/ or osdu/partition/")
        raise typer.Exit(1) # non-zero exit status

    if not policy_id.endswith(".rego"):
        policy_id = policy_id + ".rego"

    short_policy_id = os.path.basename(policy_id)

    console.print(f"Reading contents of {file} as '{policy_id}'")

    with rich.progress.open(file, "r") as f:
        data = f.read()
    f.close()

    if template:
        template_data = Template(data)
        data = template_data.substitute(
            {
            'data_partition': ctx.obj.data_partition,
            'DATA_PARTITION': ctx.obj.data_partition,
            'name': short_policy_id.removesuffix('.rego')
            }
        )
        console.print(Panel(data, title=f"rendered {policy_type} policy", highlight=True))

    if not force:
        typer.confirm(f"Add {policy_id} to OPA", abort=True)
    
    with console.status(f"Adding policy {policy_id}"):
        r = opa.put_opa_policy_direct(policy_id=policy_id, data=data, base_url=url)
    if r.ok:
        console.print(f"Policy {policy_id} added to OPA")
    else:
        error_console.print(f"Error adding policy {policy_id} to OPA {r.json()} {r.status_code}")
        raise typer.Exit(1) # non-zero exit status

@cli.command("check", rich_help_panel="Policy Developer Utils")
def opa_check(
        ctx: typer.Context,
        file_list: List[str]= typer.Argument(...),
        display: bool = typer.Option(False, "--display", "-d", help="Display policy"),
        template: bool = typer.Option(False, "--template", "-t", help="Input is a template to be rendered"),
    ):
    """
    Check rego file for errors
    """
    if file_list == ['-']:
        file_list = sys.stdin.read().strip().split()
    
    chk = opa.OpaCheck()
    for file in file_list:
        policy_id = os.path.basename(file)
        console.print(f"Reading contents of {file} as '{policy_id}'")

        with rich.progress.open(file, "r") as f:
            data = f.read()
        f.close()

        if template:
            try:
                template_data = Template(data)
                data = template_data.substitute(
                    {
                        'data_partition': ctx.obj.data_partition,
                        'DATA_PARTITION': ctx.obj.data_partition,
                        'name': policy_id.removesuffix('.rego')
                    }
                )
            except ValueError as err:
                error_console.print(f"Error in {file}: {err}")
        if display:
            console.print(Panel(data, title=f"{policy_id}", highlight=True))

        with console.status(f"Checking {policy_id}"):
            result, error_msg = chk.check(rego=data, filename=policy_id)
        if result:
            if 'rego_parse_error' in error_msg:
                error_console.print(f"{policy_id}: rego_parse_error:\n{error_msg}")
            else:
                error_console.print(f"{policy_id}: error {error_msg}")
            chk.delete() # cleanup before exit
            raise typer.Exit(1) # non-zero exit status
        else:
            console.print(f"{policy_id}: [green]OK[/]")
    chk.delete()

@cli.command("opa-rm", rich_help_panel="Policy Developer Utils")
def delete_from_opa(
        ctx: typer.Context,
        policy_list: List[str]= typer.Argument(...),
        force: bool = typer.Option(False, "--force", help="No confirmation prompt"),
        url: str = typer.Option("http://localhost:8181", help="Base URL to connect to OPA")
    ):
    """
    [red]delete[/red] a policy directly from OPA. :fire: for LOCAL testing/development

    If policy_list is - on command-line, the list will be read from standard-in
    """
    interactive = True
    if policy_list == ['-']:
        interactive = False
        policy_list = sys.stdin.read().strip().split()

    for policy_id in policy_list:

        if policy_id.startswith('osdu/instance/'):
            policy_type = "instance"
        elif policy_id.startswith('osdu/partition/'):
            policy_type = "partition"
        else:
            error_console.print("Error: policy_id must start with osdu/instance/ or osdu/partition/")
            raise typer.Exit(1) # non-zero exit status

        if not policy_id.endswith(".rego"):
            policy_id = policy_id + ".rego"
        short_policy_id = os.path.basename(policy_id)

        if not force:
            typer.confirm(f"Delete {policy_id} from OPA", abort=True)

        with console.status(f"Adding policy {policy_id}"):
            r = opa.delete_opa_policy_direct(policy_id=policy_id, base_url=url)

        if r.ok:
            console.print(f"Policy {policy_id} deleted from OPA")
        else:
            error_console.print(f"Error deleting policy {policy_id} from OPA {r.json()} {r.status_code}")
            raise typer.Exit(1) # non-zero exit status

@cli.callback()
def main(
    ctx: typer.Context,
    token: str = typer.Option(None, '-t', '--token', envvar="TOKEN"),
    policy_url: str = typer.Option(None, '-p', '--policy-url', envvar="POLICY_URL"),
    base_url: str = typer.Option(None, '--host', '--url', '--base-url', envvar="BASE_URL"),
    entitlements_url: str = typer.Option(None, '-e', '--ent-url', envvar="ENTITLEMENTS_URL", hidden=True),
    storage_url: str = typer.Option(None, '-s', '--storage-url', envvar="STORAGE_URL", hidden=True),
    search_url: str = typer.Option(None, '-x', '--ent-url', envvar="SEARCH_URL", hidden=True),
    legal_url: str = typer.Option(None, '-l', '--legal-url', envvar="LEGAL_URL", hidden=True),
    dataset_url: str = typer.Option(None, '--dataset-url', envvar="DATASET_URL", hidden=True),
    path: str = typer.Option('/api/policy/v1', envvar="POLICY_PATH", hidden=True),
    data_partition: str = typer.Option(None, '-d', '--data-partition-id', envvar="DATA_PARTITION"),
    x_user_id: str = typer.Option(None, envvar="XUSERID", help="optional user id, added to headers"),
    debug: bool = typer.Option(False, hidden=True, help="Debug output"),
    correlation_id: bool = typer.Option(True, hidden=True, help="Enable/disable correlation-id"),
    verbose: bool = typer.Option(False, "-V", "--verbose", hidden=True, help="Verbose"),
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        "-v",
        help="Show the application's version and exit.",
        callback=_version_callback,
        is_eager=True
    )
    ) -> None:

    if not token:
        error_console.print("Missing token; pass --token or set env[TOKEN]")
        raise typer.Exit(1) # non-zero exit status

    if not data_partition:
        error_console.print("Missing data_partition; pass --data-partition or set env[DATA_PARTITION]")
        raise typer.Exit(1) # non-zero exit status

    if not base_url:
        if not policy_url:
            error_console.print("Missing url and base_url; pass --url or set env[POLICY_URL]")
            raise typer.Exit(1) # non-zero exit status
        base_url = policy_url

    if not policy_url:
        policy_url = base_url

    if not entitlements_url:
        entitlements_url = base_url + '/api/entitlements/v2'

    if not storage_url:
        storage_url = base_url + '/api/storage/v2'

    if not search_url:
        search_url = base_url + '/api/search/v2'

    if not legal_url:
        legal_url = base_url + "/api/legal/v1"

    if not dataset_url:
        dataset_url = base_url + "/api/dataset/v1"

    uuid = None
    if correlation_id: 
        uuid = uuid7str()

    ctx.obj = SimpleNamespace(
            token = token,
            data_partition = data_partition,
            url = policy_url + path,
            base_url = base_url,
            diag_url = policy_url + "/diag",
            entitlements_url = entitlements_url,
            storage_url = storage_url,
            search_url = search_url,
            dataset_url = dataset_url,
            debug = debug,
            legal_url = legal_url,
            uuid = uuid,
            user_id = getpass.getuser(),
            x_user_id = x_user_id)

    if debug: # hidden option
        console.print(f"token: {ctx.obj.token}")
        console.print(f"base_url: {ctx.obj.base_url}")
        console.print(f"policy_url: {ctx.obj.url}")
        console.print(f"entitlements_url: {ctx.obj.entitlements_url}")
        console.print(f"storage_url: {ctx.obj.storage_url}")
        console.print(f"search_url: {ctx.obj.search_url}")
        console.print(f"legal_url: {ctx.obj.legal_url}")
        console.print(f"data_partition: {ctx.obj.data_partition}")
        console.print(f"dataset_url: {ctx.obj.dataset_url}") 
        console.print(f"x-user-id: {ctx.obj.x_user_id}")
        console.print(f"uuid: {ctx.obj.uuid}")
    if verbose:
        console.print(f"Correlation-ID: {ctx.obj.uuid}")

@cli.command(rich_help_panel="Policy Commands")
def add(ctx: typer.Context,
    policy_id: str = typer.Argument(...),
    #file: typer.FileBinaryRead = typer.Option(..., encoding='utf-8')
    file: Path = typer.Option(...,
        "-f", "--file",
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
    ),
    template: bool = typer.Option(False, "--template", "-t", help="Input is a template to be rendered"),
    force: bool = typer.Option(False, "--force", help="No confirmation prompt")
    ):
    """
    [green]Add or update[/green] a policy. :sparkles:

    If using option --template:

        ${data_partition} = data_partition associated with your token
        ${name} = short name you give as a policy_id on the command line
    """
    if policy_id.startswith('osdu/instance/'):
        error_console.print("Error: Instance policies not supported")
        raise typer.Abort()
    short_policy_id = os.path.basename(policy_id)
    if not short_policy_id.endswith(".rego"):
        short_policy_id = short_policy_id + ".rego"
    console.print(f"add contents of {file} as '{short_policy_id}' in {ctx.obj.data_partition}")
    with rich.progress.open(file, "r") as f:
        data = f.read()
    f.close()

    if template:
        template_data = Template(data)
        data = template_data.substitute(
            {'data_partition': ctx.obj.data_partition,
            'DATA_PARTITION': ctx.obj.data_partition,
            'name': short_policy_id.removesuffix('.rego')
        })
        console.print(Panel(data, title="rendered policy", highlight=True))

    if not force:
        typer.confirm(f"add {short_policy_id} in {ctx.obj.data_partition}", abort=True)
    bdata = data.encode('utf-8')
    files = {'file': (short_policy_id, bdata)}

    with console.status(f"Adding partition policy {short_policy_id}"):
        add_partition_policy(ctx=ctx, policy_id=short_policy_id, files=files)

@cli.command("rm", rich_help_panel="Policy Commands")
def delete(
        ctx: typer.Context,
        policy_list: List[str]= typer.Argument(...),
        force: bool = typer.Option(False, "--force", help="No confirmation prompt")
        #... if sys.stdin.isatty() else sys.stdin.read().strip()
    ):
    """
    [red]delete[/red] a partition policy. :fire:

    If policy_list is - on command-line, the list will be read from standard-in
    """
    interactive = True
    if policy_list == ['-']:
        interactive = False
        policy_list = sys.stdin.read().strip().split()

    for policy_id in policy_list:
        if "osdu/instance/" in policy_id:
            error_console.print("Error: Instance policies not supported")
            raise typer.Abort()

        short_policy_id = os.path.basename(policy_id)
        if not short_policy_id.endswith(".rego"):
            short_policy_id = short_policy_id + ".rego"

        with console.status(f"Checking partition policy {short_policy_id}"):
            result = get_partition_policy(ctx, short_policy_id)

        delete = False
        if result:
            long_policy_id = result['id']
            if not force and interactive:
                delete = typer.confirm(f"Delete {long_policy_id} in {ctx.obj.data_partition}", abort=True)
            elif not force and not interactive:
                console.print(f"preview: would have deleted '{long_policy_id} in {ctx.obj.data_partition}'. Use --force")
                continue

            if delete or force:
                console.print(f"deleting {short_policy_id}")
                with console.status(f"Deleting partition policy {short_policy_id}"):
                    result = delete_partition_policy(ctx, short_policy_id)
                if result:
                    console.print(f"Policy [green]{long_policy_id}[/] deleted")
                else:
                    error_console.print(f"Delete of {long_policy_id} failed")
        else:
            if force:
                with console.status(f"Deleting partition policy {short_policy_id}"):
                    result = delete_partition_policy(ctx, short_policy_id)
                if result:
                    console.print(f"Policy [green]{short_policy_id}[/] deleted")
                else:
                    error_console.print(f"Delete of {short_policy_id} failed")
            else:
                error_console.print(f"Skipping delete of {policy_id}")

def get_info(ctx: typer.Context, base_url):
    """
    Get info
    """
    try:
        with console.status("Getting info..."):
            r = requests.get(base_url + "/info", headers=headers(ctx))
    except requests.exceptions.RequestException as e: 
        raise SystemExit(e)

    if r.ok:
        console.print(r.json())
    else:
        error_console.print(r.text)
        raise typer.Exit(1) # non-zero exit status

@cli.command(rich_help_panel="Utils")
def storage(ctx: typer.Context,
    id_list: List[str]= typer.Argument(..., help="IDs to retrieve"),
    versions: bool = typer.Option(False, "--versions", "-V", help="Show versions"),
    version: str = typer.Option(None, "-v", "--version", help="Get particular version"),
    dataset: bool = typer.Option(False, "--dataset", "-D", help="dataset"),
    get: bool = typer.Option(False, "--get", hidden=True, help="get dataset"),
    download: bool = typer.Option(False, "--download", "-d", help="download dataset"),
    raw: bool = typer.Option(False, "--raw", help="json output"),
    ):
    """
    Storage and Dataset record retrieval utility
    """
    if id_list == ['-']:
        id_list = sys.stdin.read().strip().split()

        if not len(id_list):
            error_console.print("Error: Missing input 'ID_LIST...'")
            raise typer.Exit(1) # non-zero exit status

    for id in id_list:
        try:
            if versions:
                url = ctx.obj.storage_url + "/records/versions/" + id
            elif version:
                url = ctx.obj.storage_url + "/records/" + id + "/" + version
            elif dataset:
                url = ctx.obj.dataset_url + "/getDatasetRegistry?id=" + id
            elif get or download:
                url = ctx.obj.dataset_url + "/getRetrievalInstructions?id=" + id
                #url = ctx.obj.dataset_url + "/getRetrievalInstructions?kindSubType=" + id
            else:
                url = ctx.obj.storage_url + "/records/" + id

            with console.status(f"Retrieving record..."):
                r = requests.get(url,
                    headers=headers(ctx)
                )
        except requests.exceptions.RequestException as e: 
            raise SystemExit(e)
        
        if r.ok:
            if raw:
                console.print(r.json())
            elif get or download:
                get_dataset(ctx=ctx, data=r.json(), download=download)
            else:
                console.print(r.json())
        else:
            error_console.print("Error:", r.text, r.status_code)
            raise typer.Exit(1) # non-zero exit status

def get_filename_from_storage(ctx: typer.Context, id):
    try:
        with console.status(f"Retrieving storage record {id}..."):
            r = requests.get(ctx.obj.storage_url + "/records/" + id,
                headers=headers(ctx)
            )
    except requests.exceptions.RequestException as e: 
        raise SystemExit(e)
    
    if r.ok:
        r_json = r.json()
        if "data" in r_json and  \
            "DatasetProperties" in r_json and \
            "FileSourceInfo" in r_json and \
            "Name" in r_json:
            name = r.json()["data"]["DatasetProperties"]["FileSourceInfo"]["Name"]
            return name
        else:
            error_console.print(f"Error: FileSourceInfo Name Path missing ['data']['DatasetProperties']['FileSourceInfo']['Name'] in record {id}")
            raise typer.Exit(1) # non-zero exit status

def get_dataset(ctx: typer.Context, data, download=False, raw=False):
    id = data["delivery"][0]["datasetRegistryId"]
    unsigned_url = data["delivery"][0]["retrievalProperties"]["unsignedUrl"]
    name = get_filename_from_storage(ctx=ctx, id=id)
    filename = os.path.basename(name)
    signed_url = data["delivery"][0]["retrievalProperties"]["signedUrl"]
    
    if download:
        try:
            with console.status(f"Downloading {id} {name}..."):
                r = requests.get(signed_url)
        except requests.exceptions.RequestException as e: 
            raise SystemExit(e)

        if r.ok:
            with console.status(f"Saving {filename}..."):
                f = open(filename, "wb")
                f.write(r.content)
                f.close()
            console.print(f"Downloaded '{name}' from '{id}' as '{filename}'")
    else:
        console.print(f"unsigned_url: {unsigned_url}")
        console.print(f"signed_url: {signed_url}")

@cli.command(rich_help_panel="Utils")
def search(ctx: typer.Context,
    id: str = typer.Argument(None),
    kind: str = typer.Option("*:*:*:*", help="The kind(s) of the record to query"),
    query: str = typer.Option("", help="Query string based on Lucene query string syntax"),
    sort_field: str = typer.Option("id", help="Sort field"),
    sort_order: str = typer.Option("ASC", help="Sort Order", callback=search_cli.search_order_callback),
    limit: int = typer.Option(10, help="The maximum number of results to return from the given offset"),
    offset: int = typer.Option(0, help="The starting offset from which to return results"),
    detail: bool = typer.Option(False, "--detail", help="Show more details"),
    id_only: bool = typer.Option(False, "--id-only", help="Only show IDs"),
    random_id: bool = typer.Option(False, "--random-id", help="Get a random ID"),
    output: OutputType = typer.Option("fancy", help="Output style"),
    file: Path = typer.Option("search.xlsx",
            "-f", "--file",
            dir_okay=False,
            resolve_path=True,
            help="Excel file for output=excel"
        ),
    open_file: bool = typer.Option(False, "--open", help="Open file after saving"),
    raw: bool = typer.Option(False, "--raw", help="Show raw/json search results"),
    spatial_filter: Path = typer.Option(None,
            exists=True,
            file_okay=True,
            dir_okay=False,
            writable=False,
            readable=True,
            resolve_path=True,
            help="Spatial filter json file"
        ),
    with_cursor: bool = typer.Option(False, "--with-cursor", help="Use cursors"),
    cursor: str = typer.Option(None, "--cursor", help="Pass cursor back")
    ):
    """
    Search utility

    Kind must follow the convention: {Schema-Authority}:{dataset-name}:{record-type}:{version}
    
    Example Searches:
    Wellbore Master Data Instances for Well with ID 1691:
    --kind="*:*:master-data--Wellbore:*" --query=data.WellID:\\"osdu:master-data--Well:1691:\\"
    
    Wellbore Trajectory Work Product Components associated with Wellbore ID 1691:
    --kind=*:*:work-product-component--WellboreTrajectory:* --query=data.WellboreID:\\"osdu:master-data--Wellbore:1691:\\"

    Any record with any field equal "well":
    --kind="*:*:*:*" --query=well

    Where source is blended or TNO:
    --kind="*:*:*:*" --query="data.Source:(BLENDED TNO)"

    Where source is exactly "TNO":
    --kind="*:*:*:*" --query=data.Source:\\"TNO\\"

    All wellbore logs from 2022 year:
    --kind="*:*:work-product-component--WellLog:*" --query="createTime:[2022-01-01 TO 2022-12-31]"

    All well logs deeper than 4000m:
    --kind="*:*:work-product-component--WellLog:*" query="data.BottomMeasuredDepth:[4000 TO *]"

    All well logs deeper than 2000m or shallower than 4000m:
    --kind="*:*:work-product-component--WellLog:*" --query="data.BottomMeasuredDepth:(>=2000 OR <=4000)"

    Note: --with-cursor
    To process the next --with-cursor request, the search service keeps the search context alive for 1 minute,
    which is the time required to process the next batch of results. Each cursor request sets a new expiry time.
    The cursor will expire after 1 min and won't return any more results if the requests are not made in specified time.
    """
    if id:
        query=f"id:\"{id}\""

    # Note: above \\ is for help output, so just one \ is needed

    # search_json_data = {
    #     'kind': kind,
    #     'query': query,
    #     'offset': offset,
    #     'limit': limit,
    #     'queryAsOwner': queryasowner,
    #     'aggregateBy': aggregateby,
    #     'sort': {
    #         'field': [
    #             'id',
    #         ],
    #         'order': [
    #             'ASC',
    #         ],
    #     },
    #     'returnedFields': [],
    # }

    sort_data = {
        'field': [ sort_field ],
        'order': [ sort_order ]
    }
    search_json_data = {
        'kind': kind,
        'query': query,
        'limit': limit,
        'offset': offset,
        'sort': sort_data
    }

    if spatial_filter:
        try:
            with rich.progress.open(spatial_filter, "r") as f:
                spatial_filter_data = json.load(f)
            f.close()
        except json.decoder.JSONDecodeError as err:
            error_console.print(f"Error: {spatial_filter} malformed json. {err}")
            raise typer.Exit(1) # non-zero exit status

        search_json_data["spatialFilter"] = spatial_filter_data

    if with_cursor:
        url = ctx.obj.search_url + "/query_with_cursor"
        if cursor:
            search_json_data["cursor"] = cursor
    else:
        url = ctx.obj.search_url + "/query"

    if ctx.obj.debug:
        console.print(search_json_data)

    try:
        with console.status(f"Retrieving search results..."):
            r = requests.post(url,
                json=search_json_data,
                headers=headers(ctx)
                )
    except requests.exceptions.RequestException as e: 
        raise SystemExit(e)

    try: 
        # silly math for figuring out terminal size
        width, height = os.get_terminal_size()
    except:
        # not a tty... no worries just simple output
        output = output.simple

    if r.ok:
        if "results" in r.json():
            count=len(r.json()["results"])

            if raw:
                console.print(r.json())
            elif not count:
                error_console.print("No results found")
                raise typer.Exit(2) # non-zero exit status
            elif id_only:
                search_cli.display_search_results_idonly(r.json())
            elif random_id:
                search_cli.display_search_results_idonly_random(r.json())
            elif output == output.fancy:
                search_cli.display_search_results_fancy(r.json(), show_kind=detail)
            elif output == output.simple:
                search_cli.display_search_results_simple(r.json())
            elif output == output.excel:
                with console.status(f"Saving search results..."):
                    search_cli.display_search_results_excel(r.json(), show_kind=detail, excelfile=file.name)
                #os.system(f"open -a'Microsoft Excel.app' {file.name}")
                #os.startfile(f"{file.name}")
                if open_file:
                    search_cli.open_file(file.name)
            elif output == output.tree:
                error_console.print("Error: Not supported output type")
                raise typer.Exit(1) # non-zero exit status

        if with_cursor:
            console.print(f":warning-emoji: [bold] Cursor: [blue]{r.json()['cursor']}")
    else:
        error_console.print("Error:", r.text, r.status_code)
        raise typer.Exit(1) # non-zero exit status

if __name__ == "__main__":
    cli()
